#!/usr/bin/env python
"""
Tools for accessing (searching and downloading) datasets from the USGS

"""

import datetime
from typing import List, Dict

import rsgislib
import rsgislib.tools.utils
import rsgislib.tools.httptools

DATA_PRODUCTS = {
    "landsat_tm_c1": "5e83d08fd9932768",
    "landsat_etm_c1": "5e83a507d6aaa3db",
    "landsat_8_c1": "5e83d0b84df8d8c2",
    "landsat_tm_c2_l1": "5e83d0a0f94d7d8d",
    "landsat_etm_c2_l1": "5e83d0d0d2aaa488",
    "landsat_ot_c2_l1": "5e81f14ff4f9941c",
    "landsat_mss_c2_l1": "5e83d0e09a752cff",
    "landsat_tm_c2_l2": "5e83d11933473426",
    "landsat_etm_c2_l2": "5e83d12aada2e3c5",
    "landsat_ot_c2_l2": "5e83d14f30ea90a9",
    "sentinel_2a": "5e83a42c6eba8084",
}

USGS_WRS1 = 1
USGS_WRS2 = 2


USGS_M2M_URL = "https://m2m.cr.usgs.gov/api/api/json/stable/"

# Note to download data you need to registered for m2m access:
# https://ers.cr.usgs.gov/profile/access


def _check_usgs_response(data: Dict) -> Dict:
    """
    A function which checks the response for an error (producing an exception)
    and extracts the text output and returns it.

    :param data:
    :return:
    """

    if (
        rsgislib.tools.utils.dict_struct_does_path_exist(data, ["errorCode"])
        and data["errorCode"] != None
    ):
        raise rsgislib.RSGISPyException(
            "{} - {}".format(data["errorCode"], data["errorMessage"])
        )

    if not rsgislib.tools.utils.dict_struct_does_path_exist(data, ["data"]):
        print(data)
        raise rsgislib.RSGISPyException("Data structure is not as expected - check.")

    return data["data"]


def usgs_login(username: str = None, password: str = None) -> str:
    """
    A function to login to the USGS m2m service.

    :param username: Your username for USGS EarthExplorer. If RSGIS_USGS_USER
                     environmental variable is specified then username will
                     read from there is None is passed (Default: None)
    :param password: Your password for USGS EarthExplorer. If RSGIS_USGS_PASS
                     environmental variable is specified then password will
                     read from there is None is passed (Default: None)
    :return: the API key for the USGS session.
    """
    if username is None:
        username = rsgislib.tools.utils.get_environment_variable("RSGIS_USGS_USER")

    if password is None:
        password = rsgislib.tools.utils.get_environment_variable("RSGIS_USGS_PASS")

    login_data = {"username": username, "password": password}
    login_url = USGS_M2M_URL + "login"
    rtnd_data = rsgislib.tools.httptools.send_http_json_request(login_url, login_data)
    api_key = _check_usgs_response(rtnd_data)
    return api_key


def usgs_logout(api_key: str):
    """
    Log out of the USGS m2m system using the api_key created at login.
    :param api_key: The API key created at login to authenticate.

    """
    logout_url = USGS_M2M_URL + "logout"
    rsgislib.tools.httptools.send_http_json_request(logout_url, None, api_key)


def can_user_dwnld(api_key: str) -> bool:
    """
    Does the user logged in with the api_key have permission to download data.

    :param api_key: The API key created at login to authenticate.
    :return: boolean - True does have permission.

    """
    permissions_url = USGS_M2M_URL + "permissions"
    rtnd_data = rsgislib.tools.httptools.send_http_json_request(
        permissions_url, api_key=api_key
    )
    permissions_info = _check_usgs_response(rtnd_data)

    have_dwn_per = False
    if (permissions_info is not None) and ("download" in permissions_info):
        have_dwn_per = True
    return have_dwn_per


def can_user_order(api_key: str) -> bool:
    """
    Does the user logged in with the api_key have permission to order data.

    :param api_key: The API key created at login to authenticate.
    :return: boolean - True does have permission.
    """
    permissions_url = USGS_M2M_URL + "permissions"
    rtnd_data = rsgislib.tools.httptools.send_http_json_request(
        permissions_url, api_key=api_key
    )
    permissions_info = _check_usgs_response(rtnd_data)
    have_ord_per = False
    if (permissions_info is not None) and ("order" in permissions_info):
        have_ord_per = True
    return have_ord_per


def get_wrs_pt(
    api_key: str, row: int, path: int, grid_version: int = USGS_WRS2
) -> (float, float):
    """
    Get a point for the WRS row/path which can be used for a query.

    :param api_key: The API key created at login to authenticate.
    :param row: integer for row
    :param path: integer for path
    :param grid_version: Whether the row/path is WRS1 or WRS2. Default: WRS2.
    :return: longitude, latitude

    """
    grid2ll_url = USGS_M2M_URL + "grid2ll"
    grid2ll_data = dict()
    if grid_version == USGS_WRS1:
        grid2ll_data["gridType"] = "WRS1"
    else:
        grid2ll_data["gridType"] = "WRS2"

    grid2ll_data["path"] = path
    grid2ll_data["row"] = row
    grid2ll_data["responseShape"] = "point"

    rtnd_data = rsgislib.tools.httptools.send_http_json_request(
        grid2ll_url, grid2ll_data, api_key=api_key
    )
    grid2ll_info = _check_usgs_response(rtnd_data)

    if grid2ll_info["shape"] == "point":
        lat = grid2ll_info["coordinates"][0]["latitude"]
        lon = grid2ll_info["coordinates"][0]["longitude"]
    else:
        raise rsgislib.RSGISPyException("The returned shape was not a point.")

    return lon, lat


def get_wrs_bbox(
    api_key: str, row: int, path: int, grid_version: int = USGS_WRS2
) -> (float, float, float, float):
    """
    Get a bbox for the WRS row/path which can be used for a query.

    :param api_key: The API key created at login to authenticate.
    :param row: integer for row
    :param path: integer for path
    :param grid_version: Whether the row/path is WRS1 or WRS2. Default: WRS2.
    :return: BBOX in lon/lat (x_min, x_max, y_min, y_max)

    """
    grid2ll_url = USGS_M2M_URL + "grid2ll"
    grid2ll_data = dict()
    if grid_version == USGS_WRS1:
        grid2ll_data["gridType"] = "WRS1"
    else:
        grid2ll_data["gridType"] = "WRS2"

    grid2ll_data["path"] = path
    grid2ll_data["row"] = row
    grid2ll_data["responseShape"] = "polygon"

    rtnd_data = rsgislib.tools.httptools.send_http_json_request(
        grid2ll_url, grid2ll_data, api_key=api_key
    )
    grid2ll_info = _check_usgs_response(rtnd_data)

    lat_vals = []
    lon_vals = []

    if grid2ll_info["shape"] == "polygon":
        for coord in grid2ll_info["coordinates"]:
            lat_vals.append(coord["latitude"])
            lon_vals.append(coord["longitude"])
    else:
        raise rsgislib.RSGISPyException("The returned shape was not a polygon.")

    min_lon = min(lon_vals)
    max_lon = max(lon_vals)
    min_lat = min(lat_vals)
    max_lat = max(lat_vals)

    return min_lon, max_lon, min_lat, max_lat


def get_dataset_info(dataset: str, api_key: str) -> Dict:
    """
    Get information on a particular dataset.

    :parma dataset: the name of the dataset
    :param api_key: The API key created at login to authenticate.
    :return: dict of dataset info.

    """
    dataset_url = USGS_M2M_URL + "dataset"
    dataset_data = {"datasetName": dataset}
    rtnd_data = rsgislib.tools.httptools.send_http_json_request(
        dataset_url, dataset_data, api_key=api_key
    )
    dataset_info = _check_usgs_response(rtnd_data)
    return dataset_info


def get_earth_explorer_datasets(api_key: str) -> Dict[str, List[str]]:
    """
    Get a structured dict of datasets available on the system. Structure
    is category and then a list of datasets under each category. Note,
    subcategories have been removed and lists merged.

    :param api_key: The API key created at login to authenticate.
    :return: a dict with lists of dataset names

    """
    dataset_url = USGS_M2M_URL + "dataset-categories"
    dataset_data = {"catalog": "EE"}
    rtnd_data = rsgislib.tools.httptools.send_http_json_request(
        dataset_url, dataset_data, api_key=api_key
    )
    dataset_info = _check_usgs_response(rtnd_data)

    dataset_names = dict()
    for dataset_id in dataset_info:
        dataset_names[dataset_info[dataset_id]["categoryName"]] = list()
        if "subCategories" in dataset_info[dataset_id]:
            for sub_cat_id in dataset_info[dataset_id]["subCategories"]:
                if type(sub_cat_id) != dict:
                    for dataset_item in dataset_info[dataset_id]["subCategories"][
                        sub_cat_id
                    ]["datasets"]:
                        dataset_names[dataset_info[dataset_id]["categoryName"]].append(
                            dataset_item["datasetAlias"]
                        )
        for dataset_item in dataset_info[dataset_id]["datasets"]:
            dataset_names[dataset_info[dataset_id]["categoryName"]].append(
                dataset_item["datasetAlias"]
            )
    return dataset_names


def usgs_search(
    dataset: str,
    api_key: str,
    start_date: datetime.datetime = None,
    end_date: datetime.datetime = None,
    cloud_min: int = 0,
    cloud_max: int = None,
    pt: List = None,
    bbox: List = None,
    poly_geom: str = None,
    months: List[int] = None,
    full_meta: bool = False,
    max_n_rslts: int = None,
    start_n: int = None,
) -> (List, Dict):
    """
    A function to search for landsat imagery from the USGS.

    :param dataset: The name of the dataset to query.
    :param api_key: The API key created at login to authenticate.
    :param start_date: Start date as a datetime object. (Earlier date)
    :param end_date: End date as a datetime object. (Later date)
    :param cloud_min: Minimum cloud cover (Default: 0)
    :param cloud_max: Maximum cloud cover.
    :param bbox: (MinX, MaxX, MinY, MaxY)
    :param pt: (X, Y)
    :param poly_geom: NOT IMPLEMENTED YET!
    :param months: List of months as ints (1-12) you want to limit the search for.
    :param full_meta: Full metadata returned (Default: False)
    :param max_n_rslts: the maximum number of scenes to be returned (cannot be
                        larger than 100 - if larger than 100 then use
                        get_all_usgs_search function.
    :param start_n: The scene number to start the data retrieval from. Note
                    you probably don't want to use this parameter but use the
                    get_all_usgs_search function.
    :return: List of scenes found and Dict of meta-data for the number of scenes
             available.

    """
    if dataset not in DATA_PRODUCTS:
        all_datasets = get_earth_explorer_datasets(api_key)
        found_dataset = False
        for cat in all_datasets:
            if dataset in all_datasets[cat]:
                found_dataset = True
                break
        if not found_dataset:
            raise rsgislib.RSGISPyException(f"No not recognise dataset: {dataset}")

    if (max_n_rslts is not None) and (max_n_rslts > 100):
        raise rsgislib.RSGISPyException(
            "The maximum number of results requested should not be higher than 100."
        )

    search_data = dict()
    search_data["datasetName"] = dataset

    if end_date is None:
        end_date = datetime.datetime.now()

    scn_filter = dict()

    if start_date is not None:
        scn_filter["acquisitionFilter"] = dict()
        scn_filter["acquisitionFilter"]["start"] = "{}".format(start_date)
        scn_filter["acquisitionFilter"]["end"] = "{}".format(end_date)

    if cloud_max is not None:
        scn_filter["cloudCoverFilter"] = dict()
        scn_filter["cloudCoverFilter"]["min"] = cloud_min
        scn_filter["cloudCoverFilter"]["max"] = cloud_max

    if pt is not None:
        scn_filter["spatialFilter"] = dict()
        scn_filter["spatialFilter"]["filterType"] = "geojson"
        scn_filter["spatialFilter"]["geoJson"] = dict()
        scn_filter["spatialFilter"]["geoJson"]["type"] = "Point"
        scn_filter["spatialFilter"]["geoJson"]["coordinates"] = [pt[1], pt[0]]
    elif bbox is not None:
        print("Use bbox to filter.")
        scn_filter["spatialFilter"] = dict()
        scn_filter["spatialFilter"]["filterType"] = "mbr"
        scn_filter["spatialFilter"]["lowerLeft"] = {
            "longitude": bbox[0],
            "latitude": bbox[2],
        }
        scn_filter["spatialFilter"]["upperRight"] = {
            "longitude": bbox[1],
            "latitude": bbox[3],
        }
    elif poly_geom is not None:
        """
        scn_filter["spatialFilter"] = dict()
        scn_filter["spatialFilter"]["filterType"] = "geoJson"
        scn_filter["spatialFilter"]["geoJson"] = dict()
        scn_filter["spatialFilter"]["geoJson"]["type"] = "Polygon"
        #self["geoJson"] = GeoJson(shape)
        """
        raise rsgislib.RSGISPyException(
            "Trying to use a polygon to filter but not implemented yet."
        )

    if months is not None:
        scn_filter["seasonalFilter"] = dict()
        scn_filter["seasonalFilter"] = months

    if len(scn_filter) > 0:
        search_data["sceneFilter"] = scn_filter

    if full_meta:
        search_data["metadataType"] = "full"

    if max_n_rslts is not None:
        search_data["maxResults"] = max_n_rslts

    if start_n is not None:
        search_data["startingNumber"] = start_n

    search_url = USGS_M2M_URL + "scene-search"
    rtnd_data = rsgislib.tools.httptools.send_http_json_request(
        search_url, search_data, api_key=api_key
    )
    srch_data = _check_usgs_response(rtnd_data)

    if srch_data["totalHits"] > 0:
        scns = srch_data["results"]
    else:
        scns = None

    srch_meta = {
        "totalHits": srch_data["totalHits"],
        "startingNumber": srch_data["startingNumber"],
        "nextRecord": srch_data["nextRecord"],
        "recordsReturned": srch_data["recordsReturned"],
    }

    return scns, srch_meta


def get_all_usgs_search(
    dataset: str,
    api_key: str,
    max_n_rslts: int = 1000,
    start_date: datetime.datetime = None,
    end_date: datetime.datetime = None,
    cloud_min: int = 0,
    cloud_max: int = None,
    pt: List = None,
    bbox: List = None,
    poly_geom: str = None,
    months: List[int] = None,
    full_meta: bool = False,
) -> List:
    """
    Uses the usgs_search function to retrive multiple 'pages' of search results. So, if
    you need more than 100 scenes you can use this function to undertake the multiple
    queries required and merge the results into a single list.

    :param dataset: The name of the dataset to query.
    :param api_key: The API key created at login to authenticate.
    :param max_n_rslts: The maximum number of scenes you want returned.
    :param start_date: Start date as a datetime object. (Earlier date)
    :param end_date: End date as a datetime object. (Later date)
    :param cloud_min: Minimum cloud cover (Default: 0)
    :param cloud_max: Maximum cloud cover.
    :param bbox: (MinX, MaxX, MinY, MaxY)
    :param pt: (X, Y)
    :param poly_geom: NOT IMPLEMENTED YET!
    :param months: List of months as ints (1-12) you want to limit the search for.
    :param full_meta: Full metadata returned (Default: False)
    :return: List of scenes found through the query.

    """
    if max_n_rslts <= 100:
        scns, srch_meta = usgs_search(
            dataset,
            api_key,
            start_date,
            end_date,
            cloud_min,
            cloud_max,
            pt,
            bbox,
            poly_geom,
            months,
            full_meta,
        )
    else:
        scns, srch_meta = usgs_search(
            dataset,
            api_key,
            start_date,
            end_date,
            cloud_min,
            cloud_max,
            pt,
            bbox,
            poly_geom,
            months,
            full_meta,
        )
        tot_n_scns = srch_meta["recordsReturned"]
        total_hits = srch_meta["totalHits"]
        if max_n_rslts > total_hits:
            max_n_rslts = total_hits
        while tot_n_scns <= max_n_rslts:
            xtr_scns, srch_meta = usgs_search(
                dataset,
                api_key,
                start_date,
                end_date,
                cloud_min,
                cloud_max,
                pt,
                bbox,
                poly_geom,
                months,
                full_meta,
                start_n=srch_meta["nextRecord"],
            )
            scns += xtr_scns
            tot_n_scns += len(xtr_scns)

    if len(scns) > max_n_rslts:
        scns = scns[0:max_n_rslts]

    return scns


def get_download_ids(scns, bulk=False) -> (List, List):
    """
    A function for extracting a list of display and entity IDs from a list
    of scenes as would have been returned by from a search query.

    :param scns: a list of the scenes
    :param bulk: If True then only scenes available for bulk download will be outputted.
    :return: List of display IDs, List of Entity IDs

    """
    scn_dsp_ids = list()
    scn_ent_ids = list()
    for scn in scns:
        if scn["options"]["download"] == True:
            if bulk and (scn["options"]["bulk"] == True):
                scn_dsp_ids.append(scn["displayId"])
                scn_ent_ids.append(scn["entityId"])
            elif not bulk:
                scn_dsp_ids.append(scn["displayId"])
                scn_ent_ids.append(scn["entityId"])
    return scn_dsp_ids, scn_ent_ids


def create_scene_list(
    api_key: str,
    dataset: str,
    scn_ent_ids: List[str],
    lst_name: str,
    lst_period: str = "P1W",
) -> int:
    """
    A function which creates a list of scenes on the system which could be
    downloaded.

    ISO 8601 duration format:
    P(n)Y(n)M(n)DT(n)H(n)M(n)S

    Where:
        P is the duration designator (referred to as "period"), and is always placed
          at the beginning of the duration.
        Y is the year designator that follows the value for the number of years.
        M is the month designator that follows the value for the number of months.
        W is the week designator that follows the value for the number of weeks.
        D is the day designator that follows the value for the number of days.
        T is the time designator that precedes the time components.
        H is the hour designator that follows the value for the number of hours.
        M is the minute designator that follows the value for the number of minutes.
        S is the second designator that follows the value for the number of seconds.

    For example: "P3Y6M4DT12H30M5S" = A duration of three years, six months, four
    days, twelve hours, thirty minutes, and five seconds.

    :param api_key: The API key created at login to authenticate user.
    :param dataset: name of the dataset
    :param scn_ent_ids: list of entity IDs
    :param lst_name: a name for the list - can be anything you want but should
                     be meaningful to you.
    :param lst_period: Period the list will exist for in ISO 8601 duration format.
                       Default is P1W (i.e., 1 week).
    :return: Number of scenes added.

    """
    if dataset not in DATA_PRODUCTS:
        all_datasets = get_earth_explorer_datasets(api_key)
        found_dataset = False
        for cat in all_datasets:
            if dataset in all_datasets[cat]:
                found_dataset = True
                break
        if not found_dataset:
            raise rsgislib.RSGISPyException(f"No not recognise dataset: {dataset}")

    add_scn_info = dict()
    add_scn_info["listId"] = lst_name
    add_scn_info["datasetName"] = dataset
    add_scn_info["idField"] = "entityId"
    add_scn_info["entityIds"] = scn_ent_ids
    add_scn_info["timeToLive"] = lst_period

    scn_lst_add_url = USGS_M2M_URL + "scene-list-add"
    rtnd_data = rsgislib.tools.httptools.send_http_json_request(
        scn_lst_add_url, add_scn_info, api_key=api_key
    )
    scn_lst_add_info = _check_usgs_response(rtnd_data)
    return scn_lst_add_info


def remove_scene_list(api_key: str, lst_name: str):
    """
    A function to remove a scene list from the system.

    :param api_key: The API key created at login to authenticate user.
    :param lst_name: a name for the list. Defined by create_scene_list.

    """
    rm_scn_info = {"listId": lst_name}
    scn_lst_add_url = USGS_M2M_URL + "scene-list-remove"
    rsgislib.tools.httptools.send_http_json_request(
        scn_lst_add_url, rm_scn_info, api_key=api_key
    )


def check_dwnld_opts(
    api_key: str,
    lst_name: str,
    dataset: str,
    dwnld_filetype: str = "bundle",
    rm_lst: bool = True,
) -> List[Dict[str, str]]:
    """

    :param api_key: The API key created at login to authenticate user.
    :param lst_name: A name for the list - Defined by create_scene_list.
    :param dataset: name of the dataset
    :param dwnld_filetype: What you want to download. Options: bundle, band or all
                           Default: is bundle which will be a tar.gz with all the
                           files for the scene.
    :param rm_lst: bool specifying whether the list should be deleted
                   once the processing has finished.
    :return: returns a list of dicts with the entityId and productId.

    """
    if not can_user_dwnld(api_key):
        raise rsgislib.RSGISPyException(
            "Your user account does not have permission to download data."
        )

    dwnld_opts_params = {"listId": lst_name, "datasetName": dataset}

    dwnld_opts_url = USGS_M2M_URL + "download-options"
    rtnd_data = rsgislib.tools.httptools.send_http_json_request(
        dwnld_opts_url, dwnld_opts_params, api_key
    )
    dnwld_opt_out_info = _check_usgs_response(rtnd_data)

    # Select products
    dwlds_lst = []
    if dwnld_filetype == "bundle":
        # select bundle files
        for prod in dnwld_opt_out_info:
            if prod["bulkAvailable"]:
                dwlds_lst.append(
                    {"entityId": prod["entityId"], "productId": prod["id"]}
                )
    elif dwnld_filetype == "band":
        # select band files
        for prod in dnwld_opt_out_info:
            if (prod["secondaryDownloads"] is not None) and (
                len(prod["secondaryDownloads"]) > 0
            ):
                for sec_dwnld in prod["secondaryDownloads"]:
                    if sec_dwnld["bulkAvailable"]:
                        dwlds_lst.append(
                            {
                                "entityId": sec_dwnld["entityId"],
                                "productId": sec_dwnld["id"],
                            }
                        )
    elif dwnld_filetype == "all":
        # select all available files
        for prod in dnwld_opt_out_info:
            if prod["bulkAvailable"]:
                dwlds_lst.append(
                    {"entityId": prod["entityId"], "productId": prod["id"]}
                )
                if (prod["secondaryDownloads"] is not None) and (
                    len(prod["secondaryDownloads"]) > 0
                ):
                    for sec_dwnld in prod["secondaryDownloads"]:
                        if sec_dwnld["bulkAvailable"]:
                            dwlds_lst.append(
                                {
                                    "entityId": sec_dwnld["entityId"],
                                    "productId": sec_dwnld["id"],
                                }
                            )
    else:
        raise rsgislib.RSGISPyException(
            "dwnld_filetype is not recognised - must be either bundle, band or all."
        )

    if rm_lst:
        remove_scene_list(api_key, lst_name)

    return dwlds_lst


def request_downloads(api_key: str, dwlds_lst: str, dwnld_label: str) -> (Dict, List):
    """
    A function to request download URLs for a download list (from check_dwnld_opts)

    :param api_key: The API key created at login to authenticate user.
    :param dwlds_lst: The available download list created by check_dwnld_opts.
    :param dwnld_label: A name for the download - can be anything you want but should
                        be meaningful to you.
    :return: a dict of download IDs and URL which are ready to be downloaded and
             a list of downloads IDs for scenes which are being prepared for download.

    """
    avail_dwn_urls = dict()
    dwnld_reqst_url = USGS_M2M_URL + "download-request"
    dwnld_reqst_data = {
        "downloads": dwlds_lst,
        "label": dwnld_label,
        "returnAvailable": True,
    }
    rtnd_data = rsgislib.tools.httptools.send_http_json_request(
        dwnld_reqst_url, dwnld_reqst_data, api_key=api_key
    )
    dwnld_reqst_info = _check_usgs_response(rtnd_data)

    for dwnld_info in dwnld_reqst_info["availableDownloads"]:
        avail_dwn_urls[dwnld_info["downloadId"]] = dwnld_info["url"]

    prep_dwnld_count = len(dwnld_reqst_info["preparingDownloads"])
    prep_dwnld_ids = list()
    if prep_dwnld_count > 0:
        for dwnld_info in dwnld_reqst_info["preparingDownloads"]:
            prep_dwnld_ids.append(dwnld_info["downloadId"])

    return avail_dwn_urls, prep_dwnld_ids



#!/usr/bin/env python
"""
The classification module provides classification functionality within RSGISLib.
"""

# import the C++ extension into this level
from ._classification import *

from typing import Dict

import rsgislib


class ClassSimpleInfoObj(object):
    """
    This is a class to store the information associated within the classification.

    :param id: Output pixel value for this class
    :param file_h5: hdf5 file (from rsgislib.zonalstats.extract_zone_img_band_values_to_hdf) with the training
                   data for the class
    :param red: Red colour for visualisation (0-255)
    :param green: Green colour for visualisation (0-255)
    :param blue: Blue colour for visualisation (0-255)

    """

    def __init__(self, id=None, file_h5=None, red=None, green=None, blue=None):
        """

        :param id: Output pixel value for this class
        :param file_h5: hdf5 file (from rsgislib.zonalstats.extract_zone_img_band_values_to_hdf) with the training
               data for the class
        :param red: Red colour for visualisation (0-255)
        :param green: Green colour for visualisation (0-255)
        :param blue: Blue colour for visualisation (0-255)

        """
        self.id = id
        self.file_h5 = file_h5
        self.red = red
        self.green = green
        self.blue = blue

    def __str__(self):
        return "{}: {}, ({}, {}, {})".format(
            self.id, self.file_h5, self.red, self.green, self.blue
        )

    def __repr__(self):
        return "{}: {}, ({}, {}, {})".format(
            self.id, self.file_h5, self.red, self.green, self.blue
        )


class ClassInfoObj(object):
    """
    This is a class to store the information associated within the classification.

    :param id: Internal unique ID value for this class (must start 0 and be consecutive between the classes)
    :param out_id: External unique ID for ther class which will be used as the output image pixel value, can
                   be any integer.
    :param train_file_h5: hdf5 file (from rsgislib.zonalstats.extract_zone_img_band_values_to_hdf) with the training
                        data for the class
    :param test_file_h5: hdf5 file (from rsgislib.zonalstats.extract_zone_img_band_values_to_hdf) with the testing
                       data for the class
    :param valid_file_h5: hdf5 file (from rsgislib.zonalstats.extract_zone_img_band_values_to_hdf) with the validation
                        data for the class
    :param red: Red colour for visualisation (0-255)
    :param green: Green colour for visualisation (0-255)
    :param blue: Blue colour for visualisation (0-255)

    """

    def __init__(
        self,
        id=None,
        out_id=None,
        train_file_h5=None,
        test_file_h5=None,
        valid_file_h5=None,
        red=None,
        green=None,
        blue=None,
    ):
        """

        :param id: Internal unique ID value for this class (must start 0 and be consecutive between the classes)
        :param out_id: External unique ID for ther class which will be used as the output image pixel value,
                       can be any integer.
        :param train_file_h5: hdf5 file (from rsgislib.zonalstats.extract_zone_img_band_values_to_hdf) with the training
                            data for the class
        :param test_file_h5: hdf5 file (from rsgislib.zonalstats.extract_zone_img_band_values_to_hdf) with the testing
                           data for the class
        :param valid_file_h5: hdf5 file (from rsgislib.zonalstats.extract_zone_img_band_values_to_hdf) with the validation
                            data for the class
        :param red: Red colour for visualisation (0-255)
        :param green: Green colour for visualisation (0-255)
        :param blue: Blue colour for visualisation (0-255)

        """
        self.id = id
        self.out_id = out_id
        self.train_file_h5 = train_file_h5
        self.test_file_h5 = test_file_h5
        self.valid_file_h5 = valid_file_h5
        self.red = red
        self.green = green
        self.blue = blue

    def __str__(self):
        return "{}={}: (Train:{}, Test:{}, Valid:{}), ({}, {}, {})".format(
            self.id,
            self.out_id,
            self.train_file_h5,
            self.test_file_h5,
            self.valid_file_h5,
            self.red,
            self.green,
            self.blue,
        )

    def __repr__(self):
        return "{}={}: (Train:{}, Test:{}, Valid:{}), ({}, {}, {})".format(
            self.id,
            self.out_id,
            self.train_file_h5,
            self.test_file_h5,
            self.valid_file_h5,
            self.red,
            self.green,
            self.blue,
        )


class ClassVecSamplesInfoObj(object):
    """
    This is a class to store the information associated with the classification vector training regions.

    :param id: Unique ID for the class (will probably be the pixel value for this class)
    :param class_name: Unique name for the class.
    :param vec_file: A vector file path with the training samples
    :param vec_lyr: The vector layer name within the vecfile for the training samples.
    :param file_h5: A file path for a HDF5 file where the pixel values for these samples will be stored.
    """

    def __init__(
        self, id=None, class_name=None, vec_file=None, vec_lyr=None, file_h5=None
    ):
        """

        :param id: Unique ID for the class (will probably be the pixel value for this class)
        :param class_name: Unique name for the class.
        :param vec_file: A vector file path with the training samples
        :param vec_lyr: The vector layer name within the vec_file for the training samples.
        :param file_h5: A file path for a HDF5 file where the pixel values for these samples will be stored.

        """
        self.id = id
        self.class_name = class_name
        self.vec_file = vec_file
        self.vec_lyr = vec_lyr
        self.file_h5 = file_h5


class SamplesInfoObj(object):
    """
    This is a class to store the information associated within the classification.

    :param class_name: The name of the class
    :param class_id: Is the classification numeric ID (i.e., output pixel value)
    :param mask_img: The input image mask from which samples are taken
    :param mask_pxl_val: The pixel value within the mask for the class
    :param out_samp_img_file: Temporary file which will store the sampled pixels.
    :param num_samps: The number of samples required.
    :param samples_h5_file: File location for the HDF5 file with the input image values for training.
    :param red: for visualisation red value.
    :param green: for visualisation green value.
    :param blue: for visualisation blue value.

    """

    def __init__(
        self,
        class_name=None,
        class_id=None,
        mask_img=None,
        mask_pxl_val=None,
        out_samp_img_file=None,
        num_samps=None,
        samples_h5_file=None,
        red=None,
        green=None,
        blue=None,
    ):
        """

        :param class_name: The name of the class
        :param class_id: Is the classification numeric ID (i.e., output pixel value)
        :param mask_img: The input image mask from which samples are taken
        :param mask_pxl_val: The pixel value within the mask for the class
        :param out_samp_img_file: Temporary file which will store the sampled pixels.
        :param num_samps: The number of samples required.
        :param samples_h5_file: File location for the HDF5 file with the input image values for training.
        :param red: for visualisation red value.
        :param green: for visualisation green value.
        :param blue: for visualisation blue value.

        """
        self.class_name = class_name
        self.class_id = class_id
        self.mask_img = mask_img
        self.mask_pxl_val = mask_pxl_val
        self.out_samp_img_file = out_samp_img_file
        self.num_samps = num_samps
        self.samples_h5_file = samples_h5_file
        self.red = red
        self.green = green
        self.blue = blue


def get_class_info_dict(
    cls_smpls_info: Dict[str, ClassSimpleInfoObj], smpls_dir: str
) -> Dict[str, ClassInfoObj]:
    """
    A function which converts a dictionary of ClassSimpleInfoObj objects into
    a dictionary of ClassInfoObj objects. This is useful when get_class_training_data
    has been used to extract samples and you then want to use
    create_train_valid_test_sets to split the samples into training, validation and
    testing datasets.

    Note. the output file names for the training, validation and testing datasets
    are as defined as the basename of the input hdf5 samples with either _train,
    _valid or _test appended on the end.

    :param cls_smpls_info: A dict with the class name as the key with a
                           ClassSimpleInfoObj instance as the value.
    :param smpls_dir: the file path for the training, validation and
                      testing datasets.
    :return: A dict with the class name as the key with a ClassInfoObj
             instance as the value.

    """
    import os
    import rsgislib.tools.filetools

    cls_smpls_fnl_info = dict()
    for id, cls_name in enumerate(cls_smpls_info):
        # Get the HDF5 file name / path
        smpls_h5_file = cls_smpls_info[cls_name].file_h5
        # Get the file basename (i.e., without directory path or file extension).
        basename = rsgislib.tools.filetools.get_file_basename(smpls_h5_file)
        train_file = os.path.join(smpls_dir, f"{basename}_train.h5")
        valid_file = os.path.join(smpls_dir, f"{basename}_valid.h5")
        test_file = os.path.join(smpls_dir, f"{basename}_test.h5")

        # Note ClassInfoObj has both an id and out_id. id has to start at 0
        # and be continuous for some of the classification algorithms
        # and therefore out_id will be the output pixel value for the class.
        # Therefore, the existing id will be copied to out_id while the
        # id value will be defined by the index from the enumerate.
        cls_smpls_fnl_info[cls_name] = ClassInfoObj(
            id=id,
            out_id=cls_smpls_info[cls_name].id,
            train_file_h5=train_file,
            test_file_h5=test_file,
            valid_file_h5=valid_file,
            red=cls_smpls_info[cls_name].red,
            green=cls_smpls_info[cls_name].green,
            blue=cls_smpls_info[cls_name].blue,
        )
        print(cls_smpls_fnl_info[cls_name])

    return cls_smpls_fnl_info


def get_class_training_data(
    img_band_info: list,
    class_vec_sample_info: list,
    tmp_dir: str,
    sub_sample: int = None,
    ref_img: str = None,
) -> dict:
    """
    A function to extract training for vector regions for a given input image set.

    :param img_band_info: A list of rsgislib.imageutils.ImageBandInfo objects to
                          define the images and bands of interest.
    :param class_vec_sample_info: List of rsgislib.classification.ClassVecSamplesInfoObj
                                  objects to define the training regions.
    :param tmp_dir: A directory for temporary outputs created during the processing.
    :param sub_sample: If not None then an integer needs to be provided which takes
                       a random selection from the available samples to balance the
                       number of samples used for the classification.
    :param ref_img: A reference image which defines the area of interest, pixel size
                    etc. for the processing. If None then an image will be generated
                    using the input images but the tmp_dir needs to be defined.
    :return: dictionary of ClassSimpleInfoObj objects.

    """
    import rsgislib.tools.filetools
    import rsgislib.tools.utils
    import rsgislib.imageutils
    import rsgislib.vectorutils.createrasters
    import rsgislib.zonalstats
    import os
    import random
    import shutil

    # Get valid mask, rasterised to this
    uid_str = rsgislib.tools.utils.uid_generator()
    tmp_lcl_dir = os.path.join(tmp_dir, "get_class_training_data_{}".format(uid_str))
    if not os.path.exists(tmp_lcl_dir):
        os.makedirs(tmp_lcl_dir)

    rasterise_ref_img = ref_img
    if ref_img is None:
        rasterise_ref_img = os.path.join(tmp_lcl_dir, "ref_img_vmsk.kea")
        rsgislib.imageutils.create_valid_mask(
            img_band_info, rasterise_ref_img, "KEA", tmp_lcl_dir
        )

    class_info = dict()
    for class_sample_info in class_vec_sample_info:
        cls_basename = rsgislib.tools.filetools.get_file_basename(
            class_sample_info.file_h5
        )
        out_vec_img = os.path.join(tmp_lcl_dir, "{}_img.kea".format(cls_basename))
        rsgislib.vectorutils.createrasters.rasterise_vec_lyr(
            class_sample_info.vec_file,
            class_sample_info.vec_lyr,
            rasterise_ref_img,
            out_vec_img,
            gdalformat="KEA",
            burn_val=class_sample_info.id,
            datatype=rsgislib.TYPE_16UINT,
            att_column=None,
            use_vec_extent=False,
            thematic=True,
            no_data_val=0,
        )

        if sub_sample is not None:
            out_vec_img_subsample = os.path.join(
                tmp_lcl_dir, "{}_img_subsample.kea".format(cls_basename)
            )
            rsgislib.imageutils.perform_random_pxl_sample_in_mask(
                out_vec_img,
                out_vec_img_subsample,
                "KEA",
                mask_vals=[class_sample_info.id],
                n_samples=sub_sample,
            )
            out_vec_img = out_vec_img_subsample

        rsgislib.zonalstats.extract_zone_img_band_values_to_hdf(
            img_band_info, out_vec_img, class_sample_info.file_h5, class_sample_info.id
        )
        rand_red_val = random.randint(1, 255)
        rand_grn_val = random.randint(1, 255)
        rand_blu_val = random.randint(1, 255)
        class_info[class_sample_info.class_name] = ClassSimpleInfoObj(
            id=class_sample_info.id,
            file_h5=class_sample_info.file_h5,
            red=rand_red_val,
            green=rand_grn_val,
            blue=rand_blu_val,
        )

    shutil.rmtree(tmp_lcl_dir)
    return class_info


def get_class_training_chips_data(
    img_band_info: list,
    class_vec_sample_info: list,
    chip_h_size: int,
    tmp_dir: str,
    ref_img: str = None,
) -> dict:
    """
    A function to extract training chips (windows/regions) for vector regions for a given input image set.

    :param img_band_info: A list of rsgislib.imageutils.ImageBandInfo objects to define the images and bands of interest.
    :param class_vec_sample_info: A list of rsgislib.classification.ClassVecSamplesInfoObj objects to define the
                               training regions.
    :param chip_h_size: is half the chip size to be extracted (i.e., 10 with output image chips 21x21, 10 pixels
                        either size of the one of interest).
    :param tmp_dir: A directory for temporary outputs created during the processing.
    :param ref_img: A reference image which defines the area of interest, pixel size etc. for the processing.
                   If None then an image will be generated using the input images but the tmpdir needs to be defined.
    :return: dictionary of ClassSimpleInfoObj objects.

    """
    import rsgislib.tools.filetools
    import rsgislib.tools.utils
    import rsgislib.imageutils
    import rsgislib.vectorutils
    import os
    import random
    import shutil

    # Get valid mask, rasterised to this
    uid_str = rsgislib.tools.utils.uid_generator()
    tmp_lcl_dir = os.path.join(
        tmp_dir, "get_class_training_chips_data_{}".format(uid_str)
    )
    if not os.path.exists(tmp_lcl_dir):
        os.makedirs(tmp_lcl_dir)

    rasterise_ref_img = ref_img
    if ref_img is None:
        rasterise_ref_img = os.path.join(tmp_lcl_dir, "ref_img_vmsk.kea")
        rsgislib.imageutils.create_valid_mask(
            img_band_info, rasterise_ref_img, "KEA", tmp_lcl_dir
        )

    classInfo = dict()
    for class_sample_info in class_vec_sample_info:
        cls_basename = rsgislib.tools.filetools.get_file_basename(
            class_sample_info.file_h5
        )
        out_vec_img = os.path.join(tmp_lcl_dir, "{}_img.kea".format(cls_basename))
        rsgislib.vectorutils.createrasters.rasterise_vec_lyr(
            class_sample_info.vecfile,
            class_sample_info.veclyr,
            rasterise_ref_img,
            out_vec_img,
            gdalformat="KEA",
            burn_val=class_sample_info.id,
            datatype=rsgislib.TYPE_16UINT,
            att_column=None,
            use_vec_extent=False,
            thematic=True,
            no_data_val=0,
        )
        rsgislib.imageutils.extractChipZoneImageBandValues2HDF(
            img_band_info,
            out_vec_img,
            class_sample_info.id,
            chip_h_size,
            class_sample_info.file_h5,
        )
        rand_red_val = random.randint(1, 255)
        rand_grn_val = random.randint(1, 255)
        rand_blu_val = random.randint(1, 255)
        classInfo[class_sample_info.classname] = ClassSimpleInfoObj(
            id=class_sample_info.id,
            file_h5=class_sample_info.file_h5,
            red=rand_red_val,
            green=rand_grn_val,
            blue=rand_blu_val,
        )
    shutil.rmtree(tmp_lcl_dir)
    return classInfo


def create_train_valid_test_sets(
    cls_in_info: dict,
    cls_out_info: dict,
    test_sample: int,
    valid_sample: int,
    train_sample: int = None,
    rnd_seed: int = 42,
    datatype: int = None,
):
    """
    A function which takes a dict of rsgislib.classification.ClassSimpleInfoObj
    such as those retrieved from get_class_training_data and a dict of
    rsgislib.classification.ClassInfoObj and creates the train, test, valid
    data samples from a single input file for all the classes. This is a simple
    wrapper function around the split_sample_train_valid_test function making it
    easier to process multiple classes.

    :param cls_in_info: a dict of rsgislib.classification.ClassSimpleInfoObj objects
                        with the input HDF5 file paths which will be split.
    :param cls_out_info: a dict of rsgislib.classification.ClassInfoObj objects
                         which specifies the output paths for the output HDF5 files.
    :param test_sample: The size of the testing sample to be taken.
    :param valid_sample: The size of the validation sample to be taken.
    :param train_sample: The size of the training sample to be taken. If None then the
                         remaining samples are returned.
    :param rnd_seed: The random seed to be used to randomly select the sub-samples.
    :param datatype: is the data type used for the output HDF5 file
                     (e.g., rsgislib.TYPE_32FLOAT). If None (default)
                     then the output data type will be float32.

    """
    if cls_in_info.keys() != cls_out_info.keys():
        raise rsgislib.RSGISPyException(
            "The dict keys for the input and output info do not match."
        )

    for cls_name in cls_in_info:
        split_sample_train_valid_test(
            cls_in_info[cls_name].file_h5,
            cls_out_info[cls_name].train_file_h5,
            cls_out_info[cls_name].valid_file_h5,
            cls_out_info[cls_name].test_file_h5,
            test_sample,
            valid_sample,
            train_sample,
            rnd_seed,
            datatype,
        )


def get_num_samples(in_h5_file: str) -> int:
    """
    A function to return the number of samples within the input HDF5 file.

    :param in_h5_file: Input HDF file, probably from rsgislib.zonalstats.extract_zone_img_band_values_to_hdf.
    :return: the number of samples in the hdf5 file.
    """
    import h5py

    f = h5py.File(in_h5_file, "r")
    n_samples = f["DATA/DATA"].shape[0]
    f.close()
    return n_samples


def split_sample_train_valid_test(
    in_h5_file,
    train_h5_file,
    valid_h5_file,
    test_h5_file,
    test_sample,
    valid_sample,
    train_sample=None,
    rnd_seed=42,
    datatype=None,
):
    """
    A function to split a HDF5 samples file (from rsgislib.zonalstats.extract_zone_img_band_values_to_hdf)
    into three (i.e., Training, Validation and Testing).

    :param in_h5_file: Input HDF file, probably from rsgislib.zonalstats.extract_zone_img_band_values_to_hdf.
    :param train_h5_file: Output file with the training data samples (this has the number of samples left following
                          the removal of the test and valid samples if train_sample=None)
    :param valid_h5_file: Output file with the valid data samples.
    :param test_h5_file: Output file with the testing data samples.
    :param test_sample: The size of the testing sample to be taken.
    :param valid_sample: The size of the validation sample to be taken.
    :param train_sample: The size of the training sample to be taken. If None then the remaining samples are returned.
    :param rnd_seed: The random seed to be used to randomly select the sub-samples.
    :param datatype: is the data type used for the output HDF5 file (e.g., rsgislib.TYPE_32FLOAT). If None (default)
                     then the output data type will be float32.

    """
    import rsgislib.tools.utils
    import rsgislib.tools.filetools
    import rsgislib.zonalstats
    import os

    uid_str = rsgislib.tools.utils.uid_generator()
    out_dir = os.path.split(os.path.abspath(test_h5_file))[0]
    if datatype is None:
        datatype = rsgislib.TYPE_32FLOAT
    tmp_train_valid_sample_file = os.path.join(
        out_dir, "train_valid_tmp_sample_{}.h5".format(uid_str)
    )
    rsgislib.zonalstats.split_sample_hdf5_file(
        in_h5_file,
        test_h5_file,
        tmp_train_valid_sample_file,
        test_sample,
        rnd_seed,
        datatype,
    )
    if train_sample is not None:
        tmp_train_sample_file = os.path.join(
            out_dir, "train_tmp_sample_{}.h5".format(uid_str)
        )
        rsgislib.zonalstats.split_sample_hdf5_file(
            tmp_train_valid_sample_file,
            valid_h5_file,
            tmp_train_sample_file,
            valid_sample,
            rnd_seed,
            datatype,
        )
        tmp_remain_sample_file = os.path.join(
            out_dir, "remain_tmp_sample_{}.h5".format(uid_str)
        )
        rsgislib.zonalstats.split_sample_hdf5_file(
            tmp_train_sample_file,
            train_h5_file,
            tmp_remain_sample_file,
            train_sample,
            rnd_seed,
            datatype,
        )
        rsgislib.tools.filetools.delete_file_silent(tmp_train_sample_file)
        rsgislib.tools.filetools.delete_file_silent(tmp_remain_sample_file)
    else:
        rsgislib.zonalstats.split_sample_hdf5_file(
            tmp_train_valid_sample_file,
            valid_h5_file,
            train_h5_file,
            valid_sample,
            rnd_seed,
            datatype,
        )
    rsgislib.tools.filetools.delete_file_silent(tmp_train_valid_sample_file)


def split_chip_sample_train_valid_test(
    in_h5_file,
    train_h5_file,
    valid_h5_file,
    test_h5_file,
    test_sample,
    valid_sample,
    train_sample=None,
    rnd_seed=42,
    datatype=None,
):
    """
    A function to split a chip HDF5 samples file (from rsgislib.zonalstats.extract_chip_zone_image_band_values_to_hdf)
    into three (i.e., Training, Validation and Testing).

    :param in_h5_file: Input HDF file, probably from rsgislib.zonalstats.extract_chip_zone_image_band_values_to_hdf.
    :param train_h5_file: Output file with the training data samples (this has the number of samples left following
                          the removal of the test and valid samples if train_sample=None)
    :param valid_h5_file: Output file with the valid data samples.
    :param test_h5_file: Output file with the testing data samples.
    :param test_sample: The size of the testing sample to be taken.
    :param valid_sample: The size of the validation sample to be taken.
    :param train_sample: The size of the training sample to be taken. If None then the remaining samples are returned.
    :param rnd_seed: The random seed to be used to randomly select the sub-samples.
    :param datatype: is the data type used for the output HDF5 file (e.g., rsgislib.TYPE_32FLOAT). If None (default)
                     then the output data type will be float32.

    """
    import rsgislib.tools.utils
    import rsgislib.tools.filetools
    import rsgislib.zonalstats
    import os

    uid_str = rsgislib.tools.utils.uid_generator()
    out_dir = os.path.split(os.path.abspath(test_h5_file))[0]
    if datatype is None:
        datatype = rsgislib.TYPE_32FLOAT

    tmp_train_valid_sample_file = os.path.join(
        out_dir, "train_valid_tmp_sample_{}.h5".format(uid_str)
    )
    rsgislib.zonalstats.split_sample_chip_hdf5_file(
        in_h5_file,
        test_h5_file,
        tmp_train_valid_sample_file,
        test_sample,
        rnd_seed,
        datatype,
    )
    if train_sample is not None:
        tmp_train_sample_file = os.path.join(
            out_dir, "train_tmp_sample_{}.h5".format(uid_str)
        )
        rsgislib.zonalstats.split_sample_chip_hdf5_file(
            tmp_train_valid_sample_file,
            valid_h5_file,
            tmp_train_sample_file,
            valid_sample,
            rnd_seed,
            datatype,
        )
        tmp_remain_sample_file = os.path.join(
            out_dir, "remain_tmp_sample_{}.h5".format(uid_str)
        )
        rsgislib.zonalstats.split_sample_chip_hdf5_file(
            tmp_train_sample_file,
            train_h5_file,
            tmp_remain_sample_file,
            train_sample,
            rnd_seed,
            datatype,
        )
        rsgislib.tools.filetools.delete_file_silent(tmp_train_sample_file)
        rsgislib.tools.filetools.delete_file_silent(tmp_remain_sample_file)
    else:
        rsgislib.zonalstats.split_sample_chip_hdf5_file(
            tmp_train_valid_sample_file,
            valid_h5_file,
            train_h5_file,
            valid_sample,
            rnd_seed,
            datatype,
        )
    rsgislib.tools.filetools.delete_file_silent(tmp_train_valid_sample_file)


def split_chip_sample_ref_train_valid_test(
    in_h5_file,
    train_h5_file,
    valid_h5_file,
    test_h5_file,
    test_sample,
    valid_sample,
    train_sample=None,
    rnd_seed=42,
    datatype=None,
):
    """
    A function to split a chip HDF5 samples file (from rsgislib.zonalstats.extract_chip_zone_image_band_values_to_hdf)
    into three (i.e., Training, Validation and Testing).

    :param in_h5_file: Input HDF file, probably from rsgislib.zonalstats.extract_chip_zone_image_band_values_to_hdf.
    :param train_h5_file: Output file with the training data samples (this has the number of samples left following
                          the removal of the test and valid samples if train_sample=None)
    :param valid_h5_file: Output file with the valid data samples.
    :param test_h5_file: Output file with the testing data samples.
    :param test_sample: The size of the testing sample to be taken.
    :param valid_sample: The size of the validation sample to be taken.
    :param train_sample: The size of the training sample to be taken. If None then the remaining samples are returned.
    :param rnd_seed: The random seed to be used to randomly select the sub-samples.
    :param datatype: is the data type used for the output HDF5 file (e.g., rsgislib.TYPE_32FLOAT). If None (default)
                     then the output data type will be float32.

    """
    import rsgislib
    import rsgislib.tools.filetools
    import rsgislib.tools.utils
    from rsgislib.zonalstats import split_sample_ref_chip_hdf5_file
    import os

    uid_str = rsgislib.tools.utils.uid_generator()
    out_dir = os.path.split(os.path.abspath(test_h5_file))[0]
    if datatype is None:
        datatype = rsgislib.TYPE_32FLOAT

    tmp_train_valid_sample_file = os.path.join(
        out_dir, "train_valid_tmp_sample_{}.h5".format(uid_str)
    )
    split_sample_ref_chip_hdf5_file(
        in_h5_file,
        test_h5_file,
        tmp_train_valid_sample_file,
        test_sample,
        rnd_seed,
        datatype,
    )
    if train_sample is not None:
        tmp_train_sample_file = os.path.join(
            out_dir, "train_tmp_sample_{}.h5".format(uid_str)
        )
        split_sample_ref_chip_hdf5_file(
            tmp_train_valid_sample_file,
            valid_h5_file,
            tmp_train_sample_file,
            valid_sample,
            rnd_seed,
            datatype,
        )
        tmp_remain_sample_file = os.path.join(
            out_dir, "remain_tmp_sample_{}.h5".format(uid_str)
        )
        split_sample_ref_chip_hdf5_file(
            tmp_train_sample_file,
            train_h5_file,
            tmp_remain_sample_file,
            train_sample,
            rnd_seed,
            datatype,
        )
        rsgislib.tools.filetools.delete_file_silent(tmp_train_sample_file)
        rsgislib.tools.filetools.delete_file_silent(tmp_remain_sample_file)
    else:
        split_sample_ref_chip_hdf5_file(
            tmp_train_valid_sample_file,
            valid_h5_file,
            train_h5_file,
            valid_sample,
            rnd_seed,
            datatype,
        )
    rsgislib.tools.filetools.delete_file_silent(tmp_train_valid_sample_file)


def flip_chip_hdf5_file(input_h5_file, output_h5_file, datatype=None):
    """
    A function which flips each sample in both the x and y axis. So the
    output file will have double the number of samples as the input file.

    :param input_h5_file: The input HDF5 file for chips extracted from images.
    :param output_h5_file: The output HDF5 file for chips extracted from images.
    :param datatype: is the data type used for the output HDF5 file (e.g., rsgislib.TYPE_32FLOAT). If None (default)
                     then the output data type will be float32.

    """
    import tqdm
    import h5py
    import numpy
    import rsgislib

    if datatype is None:
        datatype = rsgislib.TYPE_32FLOAT

    f = h5py.File(input_h5_file, "r")
    n_in_feats = f["DATA/DATA"].shape[0]
    chip_size = f["DATA/DATA"].shape[1]
    n_bands = f["DATA/DATA"].shape[3]

    n_out_feats = n_in_feats * 2

    feat_arr = numpy.zeros(
        [n_out_feats, chip_size, chip_size, n_bands], dtype=numpy.float32
    )

    i_feat = 0
    for n in tqdm.tqdm(range(n_in_feats)):
        numpy.copyto(
            feat_arr[i_feat], numpy.flip(f["DATA/DATA"][n].T, axis=1).T, casting="safe"
        )
        i_feat += 1
        numpy.copyto(
            feat_arr[i_feat], numpy.flip(f["DATA/DATA"][n].T, axis=2).T, casting="safe"
        )
        i_feat += 1

    f.close()

    ######################################################################
    # Create the output HDF5 file and populate with data.
    ######################################################################
    # Chunk size needs to be less than number of features
    if n_in_feats < 250:
        chunk_features = n_in_feats
    else:
        chunk_features = 250
    h5_dtype = rsgislib.get_numpy_char_codes_datatype(datatype)
    fH5Out = h5py.File(output_h5_file, "w")
    dataGrp = fH5Out.create_group("DATA")
    metaGrp = fH5Out.create_group("META-DATA")
    dataGrp.create_dataset(
        "DATA",
        data=feat_arr,
        chunks=(chunk_features, chip_size, chip_size, n_bands),
        compression="gzip",
        shuffle=True,
        dtype=h5_dtype,
    )
    describDS = metaGrp.create_dataset("DESCRIPTION", (1,), dtype="S10")
    describDS[0] = "IMAGE REF TILES".encode()
    fH5Out.close()
    ######################################################################


def flip_ref_chip_hdf5_file(input_h5_file, output_h5_file, datatype=None):
    """
    A function which flips each sample in both the x and y axis. So the
    output file will have double the number of samples as the input file.

    :param input_h5_file: The input HDF5 file for chips extracted from images.
    :param output_h5_file: The output HDF5 file for chips extracted from images.
    :param datatype: is the data type used for the output HDF5 file (e.g.,
                     rsgislib.TYPE_32FLOAT). If None (default) then the output
                     data type will be float32.

    """
    import tqdm
    import h5py
    import numpy
    import rsgislib

    if datatype is None:
        datatype = rsgislib.TYPE_32FLOAT

    f = h5py.File(input_h5_file, "r")
    n_in_feats = f["DATA/REF"].shape[0]
    chip_size = f["DATA/REF"].shape[1]
    n_bands = f["DATA/DATA"].shape[3]

    n_out_feats = n_in_feats * 2

    feat_arr = numpy.zeros(
        [n_out_feats, chip_size, chip_size, n_bands], dtype=numpy.float32
    )
    feat_ref_arr = numpy.zeros([n_out_feats, chip_size, chip_size], dtype=numpy.uint16)

    i_feat = 0
    for n in tqdm.tqdm(range(n_in_feats)):
        numpy.copyto(
            feat_ref_arr[i_feat],
            numpy.flip(f["DATA/REF"][n].T, axis=0).T,
            casting="safe",
        )
        numpy.copyto(
            feat_arr[i_feat], numpy.flip(f["DATA/DATA"][n].T, axis=1).T, casting="safe"
        )
        i_feat += 1
        numpy.copyto(
            feat_ref_arr[i_feat],
            numpy.flip(f["DATA/REF"][n].T, axis=1).T,
            casting="safe",
        )
        numpy.copyto(
            feat_arr[i_feat], numpy.flip(f["DATA/DATA"][n].T, axis=2).T, casting="safe"
        )
        i_feat += 1

    f.close()

    ######################################################################
    # Create the output HDF5 file and populate with data.
    ######################################################################
    # Chunk size needs to be less than number of features
    if n_in_feats < 250:
        chunk_features = n_in_feats
    else:
        chunk_features = 250
    h5_dtype = rsgislib.get_numpy_char_codes_datatype(datatype)
    fH5Out = h5py.File(output_h5_file, "w")
    dataGrp = fH5Out.create_group("DATA")
    metaGrp = fH5Out.create_group("META-DATA")
    dataGrp.create_dataset(
        "DATA",
        data=feat_arr,
        chunks=(chunk_features, chip_size, chip_size, n_bands),
        compression="gzip",
        shuffle=True,
        dtype=h5_dtype,
    )
    dataGrp.create_dataset(
        "REF",
        data=feat_ref_arr,
        chunks=(chunk_features, chip_size, chip_size),
        compression="gzip",
        shuffle=True,
        dtype="H",
    )
    describDS = metaGrp.create_dataset("DESCRIPTION", (1,), dtype="S10")
    describDS[0] = "IMAGE REF TILES".encode()
    fH5Out.close()
    ######################################################################


def label_pxl_sample_chips(
    sample_pxls_img,
    cls_msk_img,
    out_image,
    gdalformat,
    chip_size,
    cls_lut,
    sample_pxl_img_band=1,
    cls_msk_img_band=1,
):
    """
    A function which labels image pixels based on the proportions of a class within a chip around the
    pixel (can be used in combination with rsgislib.imageutils.assign_random_pxls). It is expected that
    this function will be used when trying to use existing maps to create deep learning chip classification
    training data.

    Pixels are labelled if the proportion of pixels is >= the threshold provided in the LUT. If more than
    one class meets the threshold then the one with the highest proportion is assigned.

    :param sample_pxls_img: The input binary image with the pixel locations (value == 1)
    :param cls_msk_img: The classification image used to assign the output pixel values.
    :param out_image: The output image. Single pixels with the class value will be outputted.
    :param gdalformat: The output image file format.
    :param chip_size: The size of the chip used to identify the class - would probably correspond
                      to the chip size being used for the deep learning classification. Areas used
                      is half the chip size around the pixel (i.e., the pixel from the samples image
                      will be at the centre of the chip).
    :param cls_lut: A dict look up table (LUT) with the thresholds per class for the pixel to be
                    classified as that class.
    :param sample_pxl_img_band: Default 1. The image band in the sample image.
    :param cls_msk_img_band: Default 1. The image band in the sample image.

    .. code:: python

        sample_pxls_img = 'LS5TM_20000108_latn531lonw37_r23p204_osgb_samples.kea'
        cls_msk_img = 'LS5TM_20000108_latn531lonw37_r23p204_osgb_clouds_up.kea'
        output_image = 'LS5TM_20000108_latn531lonw37_r23p204_osgb_samples_lbld.kea'

        cls_lut = dict()
        cls_lut[1] = 0.2
        cls_lut[2] = 0.2
        cls_lut[3] = 0.99

        label_pxl_sample_chips(sample_pxls_img, cls_msk_img, output_image, 'KEA', 21, cls_lut)

    """
    import rsgislib.rastergis
    from rios.imagereader import ImageReader
    from rios.imagewriter import ImageWriter
    import tqdm
    import numpy
    import math

    chip_size_odd = False
    if (chip_size % 2) != 0:
        chip_size_odd = True

    img_win_h_size = math.floor(chip_size / 2)
    img_win_size = chip_size
    n_pxls = img_win_size * img_win_size

    inImgs = list()
    inImgBands = list()
    inImgs.append(sample_pxls_img)
    inImgBands.append([sample_pxl_img_band])
    inImgs.append(cls_msk_img)
    inImgBands.append([cls_msk_img_band])

    writer = None
    reader = ImageReader(
        inImgs,
        windowxsize=200,
        windowysize=200,
        overlap=img_win_h_size,
        layerselection=inImgBands,
    )
    for (info, block) in tqdm.tqdm(reader):
        samples_msk_arr = block[0]
        blk_shp = samples_msk_arr.shape

        xSize = blk_shp[2] - (img_win_h_size * 2)
        ySize = blk_shp[1] - (img_win_h_size * 2)
        xRange = numpy.arange(img_win_h_size, img_win_h_size + xSize, 1)
        yRange = numpy.arange(img_win_h_size, img_win_h_size + ySize, 1)
        out_samp_arr = numpy.zeros_like(samples_msk_arr, dtype=numpy.uint8)
        for y in yRange:
            yMin = y - img_win_h_size
            yMax = y + img_win_h_size
            if chip_size_odd:
                yMax += 1
            for x in xRange:
                xMin = x - img_win_h_size
                xMax = x + img_win_h_size
                if chip_size_odd:
                    xMax += 1
                if samples_msk_arr[0][y][x] == 1:
                    img_blk = block[1][0, yMin:yMax, xMin:xMax]
                    uniq_vals, uniq_counts = numpy.unique(img_blk, return_counts=True)
                    uniq_dict = dict(zip(uniq_vals, uniq_counts))
                    first = True
                    for val in uniq_vals:
                        if val in cls_lut:
                            val_prop = uniq_dict[val] / n_pxls
                            if val_prop >= cls_lut[val]:
                                if first:
                                    max_val = val
                                    max_val_prop = val_prop
                                    first = False
                                elif val_prop > max_val_prop:
                                    max_val = val
                                    max_val_prop = val_prop
                    if not first:
                        out_samp_arr[0][y][x] = max_val

        if writer is None:
            writer = ImageWriter(
                out_image, info=info, firstblock=out_samp_arr, drivername=gdalformat
            )
        else:
            writer.write(out_samp_arr)
    writer.close(calcStats=False)
    rsgislib.rastergis.pop_rat_img_stats(out_image, True, True, True)


def plot_train_data(
    cls1_h5_file,
    cls2_h5_file,
    out_plots_dir,
    cls1_name="Class 1",
    cls2_name="Class 2",
    var_names=None,
):
    """
    A function which plots the training data (in HDF5 format) for two classes with histograms for the
    two axis'. Note, this plot only works for training extracted for pixel or clumps and not chip based
    training.

    This function uses the plotly library (https://plotly.com). It saves the plots to disk as PNGs so
    the plotly-orca package is also required.

    :param cls1_h5_file: Input HDF5 file with the training for class 1.
    :param cls2_h5_file: Input HDF5 file with the training for class 2.
    :param out_plots_dir: Output directory for the plots
    :param cls1_name: The name of the first class (Optional, default is 'Class 1')
    :param cls2_name: The name of the second class (Optional, default is 'Class 2')
    :param var_names: An optional list of variable names for the training. Optional,
                      otherwise call 'Var #1', 'Var #2' ... 'Var #N' etc.

    """
    import plotly.express as px
    import pandas
    import h5py
    import tqdm
    import os
    import numpy
    import rsgislib.tools.utils

    if not os.path.exists(out_plots_dir):
        raise rsgislib.RSGISPyException("The output directory does not exist")

    cls1_h5_obj = h5py.File(cls1_h5_file, "r")
    cls2_h5_obj = h5py.File(cls2_h5_file, "r")

    cls1_data = numpy.array(cls1_h5_obj["DATA/DATA"])
    cls2_data = numpy.array(cls2_h5_obj["DATA/DATA"])

    cls1_n = cls1_data.shape[0]
    cls2_n = cls2_data.shape[0]

    cls1_n_vars = cls1_data.shape[1]
    cls2_n_vars = cls2_data.shape[1]

    if cls1_n_vars != cls2_n_vars:
        raise rsgislib.RSGISPyException(
            "The number of variables must be the same for the two classes."
        )

    if var_names is not None:
        if len(var_names) != cls1_n_vars:
            raise rsgislib.RSGISPyException(
                "The number of variable names provided is not the same as "
                "the number of variables within the HDF5 files."
            )
    else:
        var_names = list()
        for var_n in range(cls1_n_vars):
            var_names.append("Var #{}".format(var_n))

    var_file_names = dict()
    for var_name in var_names:
        var_file_names[var_name] = rsgislib.tools.utils.check_str(
            var_name, rm_non_ascii=True, rm_dashs=True, rm_spaces=True, rm_punc=True
        )

    cls1_data_name = numpy.empty(cls1_n, dtype=numpy.dtype("U255"))
    cls1_data_name[...] = cls1_name

    cls2_data_name = numpy.empty(cls2_n, dtype=numpy.dtype("U255"))
    cls2_data_name[...] = cls2_name

    cls_data = numpy.concatenate([cls1_data, cls2_data])
    cls_data_name = numpy.concatenate([cls1_data_name, cls2_data_name])

    df_data = dict()
    for var_n in range(cls1_n_vars):
        df_data[var_names[var_n]] = cls_data[..., var_n]
    df_data["ClassName"] = cls_data_name

    df = pandas.DataFrame(df_data)

    for var1 in tqdm.tqdm(var_names):
        for var2 in var_names:
            out_title = "{} verses {}".format(var1, var2)
            out_file_name = "{}_v_{}.png".format(
                var_file_names[var1], var_file_names[var2]
            )
            out_plot_file = os.path.join(out_plots_dir, out_file_name)

            fig = px.scatter(
                df_data,
                x=var1,
                y=var2,
                color="ClassName",
                marginal_y="histogram",
                marginal_x="histogram",
                title=out_title,
            )

            fig.update_layout(plot_bgcolor="white")
            fig.update_layout(width=1000, height=1000)
            fig.update_xaxes(
                showline=True,
                linewidth=1,
                linecolor="black",
                ticks="inside",
                gridwidth=1,
                gridcolor="Grey",
                mirror=True,
            )
            fig.update_yaxes(
                showline=True,
                linewidth=1,
                linecolor="black",
                ticks="inside",
                col=1,
                gridwidth=1,
                gridcolor="Grey",
                mirror=True,
            )
            fig.write_image(out_plot_file)


def create_acc_pt_sets(
    vec_file: str,
    vec_lyr: str,
    out_vec_file_base: str,
    out_vec_lyr: str,
    cls_col: str,
    n_sets: int,
    sets_col: str = "set_id",
    out_format: str = "GeoJSON",
    out_ext: str = "geojson",
    shuffle_rows: bool = True,
    rnd_seed: int = None,
):
    """
    A function which splits a vector layer into n_sets where a 'class' column is used
    to ensure that there are the same number of samples per 'class' within each set.
    An example of where this function might be used is to split a set of accuracy
    assessment point for assessing the classification accuracy into multiple sets.
    Note, the output vector layers

    :param vec_file: Input vector file/path
    :param vec_lyr: Input vector layer name
    :param out_vec_file_base: The output vector file base name and path. Note,
                              the output file name will be: base{n_set}.out_ext.
                              If you want a character (e.g., underscore) between the
                              basename and the set number then include in the basename.
                              Example, out/path/vec_file_name_
    :param out_vec_lyr: the output vector layer name. The same layer name is used for
                        all the output files.
    :param cls_col: The column in the vector file which has values for the classes
    :param n_sets: The number of sets you want the input vector sorted into.
    :param sets_col: A column added to the output files with an integer representing
                     the set the row belongs to so if vector files are merged again
                     then the set information is not lost. (Default: 'set_id')
    :param out_format: The output vector file format (Default: GeoJSON)
    :param out_ext: the output vector file format extension (Default: gpkg)
    :param shuffle_rows: Boolean specifying whether the vector layer rows should be
                         shuffled before splitting into sets (Default: True)
    :param rnd_seed: If shuffling the rows then this random seed can be used to ensure
                     the shuffling is the same between runs.

    """
    import geopandas
    import math
    import numpy
    import tqdm

    # Read input vector file.
    base_gpdf = geopandas.read_file(vec_file, layer=vec_lyr)

    if cls_col not in base_gpdf.columns:
        raise rsgislib.RSGISPyException(
            "The specified class column is not within the input vector layer"
        )

    if shuffle_rows:
        base_gpdf = base_gpdf.sample(frac=1, random_state=rnd_seed).reset_index(
            drop=True
        )
    base_gpdf[sets_col] = numpy.zeros((base_gpdf.shape[0]), dtype=numpy.uint16)

    unq_cls_info = base_gpdf[cls_col].value_counts()

    sets = numpy.arange(1, n_sets + 1)

    for cls, n_vals in tqdm.tqdm(zip(unq_cls_info.index, unq_cls_info)):
        n_vals_per_set = math.floor(n_vals / n_sets)
        sets_arr = numpy.zeros(n_vals, dtype=numpy.uint8)
        s_idx = 0
        e_idx = n_vals_per_set
        for set in sets:
            sets_arr[s_idx:e_idx] = set
            s_idx = e_idx
            e_idx = e_idx + n_vals_per_set
        cls_msk = base_gpdf[cls_col] == cls
        base_gpdf.loc[cls_msk, sets_col] = sets_arr

    for set in tqdm.tqdm(sets):
        out_vec_file = "{}{}.{}".format(out_vec_file_base, set, out_ext)
        set_msk = base_gpdf[sets_col] == set
        tmp_gpdf = base_gpdf[set_msk]

        if out_format == "GPKG":
            tmp_gpdf.to_file(out_vec_file, layer=out_vec_lyr, driver=out_format)
        else:
            tmp_gpdf.to_file(out_vec_file, driver=out_format)



import os
from shutil import copy2
import pytest

H5PY_NOT_AVAIL = False
try:
    import h5py
except ImportError:
    H5PY_NOT_AVAIL = True

SKLEARN_NOT_AVAIL = False
try:
    import sklearn
except ImportError:
    SKLEARN_NOT_AVAIL = True

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
CLASSIFICATION_DATA_DIR = os.path.join(DATA_DIR, "classification")


@pytest.mark.skipif(
    (H5PY_NOT_AVAIL or SKLEARN_NOT_AVAIL),
    reason="h5py or scikit-learn dependencies not available",
)
def test_perform_sklearn_classifier_param_search():
    import rsgislib.classification
    import rsgislib.classification.classsklearn
    from sklearn.model_selection import GridSearchCV
    from sklearn.ensemble import RandomForestClassifier

    grid_search = GridSearchCV(
        RandomForestClassifier(), {"n_estimators": [10, 20, 30], "max_depth": [2, 3, 4]}
    )

    cls_info_dict = dict()
    cls_info_dict["Forest"] = rsgislib.classification.ClassInfoObj(
        id=0,
        out_id=1,
        train_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_forest_smpls_bal_train.h5"
        ),
        test_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_forest_smpls_bal_test.h5"
        ),
        valid_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_forest_smpls_bal_valid.h5"
        ),
        red=120,
        green=120,
        blue=120,
    )
    cls_info_dict["Grass"] = rsgislib.classification.ClassInfoObj(
        id=1,
        out_id=2,
        train_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_grass_smpls_bal_train.h5"
        ),
        test_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_grass_smpls_bal_test.h5"
        ),
        valid_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_grass_smpls_bal_valid.h5"
        ),
        red=120,
        green=120,
        blue=120,
    )
    cls_info_dict["Urban"] = rsgislib.classification.ClassInfoObj(
        id=2,
        out_id=3,
        train_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_urban_smpls_bal_train.h5"
        ),
        test_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_urban_smpls_bal_test.h5"
        ),
        valid_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_urban_smpls_bal_valid.h5"
        ),
        red=120,
        green=120,
        blue=120,
    )
    cls_info_dict["Water"] = rsgislib.classification.ClassInfoObj(
        id=3,
        out_id=4,
        train_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_water_smpls_bal_train.h5"
        ),
        test_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_water_smpls_bal_test.h5"
        ),
        valid_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_water_smpls_bal_valid.h5"
        ),
        red=120,
        green=120,
        blue=120,
    )

    rsgislib.classification.classsklearn.perform_sklearn_classifier_param_search(
        cls_info_dict, search_obj=grid_search
    )


@pytest.mark.skipif(
    (H5PY_NOT_AVAIL or SKLEARN_NOT_AVAIL),
    reason="h5py or scikit-learn dependencies not available",
)
def test_train_sklearn_classifier():
    import rsgislib.classification
    import rsgislib.classification.classsklearn
    from sklearn.ensemble import RandomForestClassifier

    cls_info_dict = dict()
    cls_info_dict["Forest"] = rsgislib.classification.ClassInfoObj(
        id=0,
        out_id=1,
        train_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_forest_smpls_bal_train.h5"
        ),
        test_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_forest_smpls_bal_test.h5"
        ),
        valid_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_forest_smpls_bal_valid.h5"
        ),
        red=120,
        green=120,
        blue=120,
    )
    cls_info_dict["Grass"] = rsgislib.classification.ClassInfoObj(
        id=1,
        out_id=2,
        train_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_grass_smpls_bal_train.h5"
        ),
        test_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_grass_smpls_bal_test.h5"
        ),
        valid_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_grass_smpls_bal_valid.h5"
        ),
        red=120,
        green=120,
        blue=120,
    )
    cls_info_dict["Urban"] = rsgislib.classification.ClassInfoObj(
        id=2,
        out_id=3,
        train_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_urban_smpls_bal_train.h5"
        ),
        test_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_urban_smpls_bal_test.h5"
        ),
        valid_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_urban_smpls_bal_valid.h5"
        ),
        red=120,
        green=120,
        blue=120,
    )
    cls_info_dict["Water"] = rsgislib.classification.ClassInfoObj(
        id=3,
        out_id=4,
        train_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_water_smpls_bal_train.h5"
        ),
        test_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_water_smpls_bal_test.h5"
        ),
        valid_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_water_smpls_bal_valid.h5"
        ),
        red=120,
        green=120,
        blue=120,
    )

    sk_classifier = RandomForestClassifier()

    rsgislib.classification.classsklearn.train_sklearn_classifier(
        cls_info_dict, sk_classifier
    )


@pytest.mark.skipif(
    (H5PY_NOT_AVAIL or SKLEARN_NOT_AVAIL),
    reason="h5py or scikit-learn dependencies not available",
)
def test_apply_sklearn_classifier(tmp_path):
    import rsgislib.imageutils
    import rsgislib.classification
    import rsgislib.classification.classsklearn
    from sklearn.ensemble import RandomForestClassifier

    cls_info_dict = dict()
    cls_info_dict["Forest"] = rsgislib.classification.ClassInfoObj(
        id=0,
        out_id=1,
        train_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_forest_smpls_bal_train.h5"
        ),
        test_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_forest_smpls_bal_test.h5"
        ),
        valid_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_forest_smpls_bal_valid.h5"
        ),
        red=120,
        green=120,
        blue=120,
    )
    cls_info_dict["Grass"] = rsgislib.classification.ClassInfoObj(
        id=1,
        out_id=2,
        train_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_grass_smpls_bal_train.h5"
        ),
        test_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_grass_smpls_bal_test.h5"
        ),
        valid_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_grass_smpls_bal_valid.h5"
        ),
        red=120,
        green=120,
        blue=120,
    )
    cls_info_dict["Urban"] = rsgislib.classification.ClassInfoObj(
        id=2,
        out_id=3,
        train_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_urban_smpls_bal_train.h5"
        ),
        test_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_urban_smpls_bal_test.h5"
        ),
        valid_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_urban_smpls_bal_valid.h5"
        ),
        red=120,
        green=120,
        blue=120,
    )
    cls_info_dict["Water"] = rsgislib.classification.ClassInfoObj(
        id=3,
        out_id=4,
        train_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_water_smpls_bal_train.h5"
        ),
        test_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_water_smpls_bal_test.h5"
        ),
        valid_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_water_smpls_bal_valid.h5"
        ),
        red=120,
        green=120,
        blue=120,
    )

    s2_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    s2_vld_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_vldmsk.kea")

    img_band_info = []
    img_band_info.append(
        rsgislib.imageutils.ImageBandInfo(s2_img, "s2", [1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    )

    sk_classifier = RandomForestClassifier()
    rsgislib.classification.classsklearn.train_sklearn_classifier(
        cls_info_dict, sk_classifier
    )

    output_img = os.path.join(tmp_path, "out_cls_img.kea")
    out_score_img = os.path.join(tmp_path, "out_cls_scr_img.kea")
    rsgislib.classification.classsklearn.apply_sklearn_classifier(
        cls_info_dict,
        sk_classifier,
        s2_vld_img,
        1,
        img_band_info,
        output_img,
        "KEA",
        class_clr_names=True,
        out_score_img=out_score_img,
    )

    assert os.path.exists(output_img) and os.path.exists(out_score_img)


@pytest.mark.skipif(
    (H5PY_NOT_AVAIL or SKLEARN_NOT_AVAIL),
    reason="h5py or scikit-learn dependencies not available",
)
def test_apply_sklearn_classifier_rat(tmp_path):
    import rsgislib.classification.classsklearn
    import rsgislib.rastergis
    from sklearn.ensemble import RandomForestClassifier

    ref_clumps_img = os.path.join(
        CLASSIFICATION_DATA_DIR, "sen2_20210527_aber_clumps_s2means_cls.kea"
    )
    clumps_img = os.path.join(tmp_path, "sen2_20210527_aber_clumps_s2means_cls.kea")
    copy2(ref_clumps_img, clumps_img)

    cls_info_dict = dict()
    cls_info_dict["Forest"] = rsgislib.classification.ClassInfoObj(
        id=0,
        out_id=1,
        train_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_forest_smpls_bal_train.h5"
        ),
        test_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_forest_smpls_bal_test.h5"
        ),
        valid_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_forest_smpls_bal_valid.h5"
        ),
        red=120,
        green=120,
        blue=120,
    )
    cls_info_dict["Grass"] = rsgislib.classification.ClassInfoObj(
        id=1,
        out_id=2,
        train_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_grass_smpls_bal_train.h5"
        ),
        test_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_grass_smpls_bal_test.h5"
        ),
        valid_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_grass_smpls_bal_valid.h5"
        ),
        red=120,
        green=120,
        blue=120,
    )
    cls_info_dict["Urban"] = rsgislib.classification.ClassInfoObj(
        id=2,
        out_id=3,
        train_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_urban_smpls_bal_train.h5"
        ),
        test_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_urban_smpls_bal_test.h5"
        ),
        valid_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_urban_smpls_bal_valid.h5"
        ),
        red=120,
        green=120,
        blue=120,
    )
    cls_info_dict["Water"] = rsgislib.classification.ClassInfoObj(
        id=3,
        out_id=4,
        train_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_water_smpls_bal_train.h5"
        ),
        test_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_water_smpls_bal_test.h5"
        ),
        valid_file_h5=os.path.join(
            CLASSIFICATION_DATA_DIR, "cls_water_smpls_bal_valid.h5"
        ),
        red=120,
        green=120,
        blue=120,
    )

    sk_classifier = RandomForestClassifier()

    rsgislib.classification.classsklearn.train_sklearn_classifier(
        cls_info_dict, sk_classifier
    )

    variables = [
        "b1Mean",
        "b2Mean",
        "b3Mean",
        "b4Mean",
        "b5Mean",
        "b6Mean",
        "b7Mean",
        "b8Mean",
        "b9Mean",
        "b10Mean",
    ]
    rsgislib.classification.classsklearn.apply_sklearn_classifier_rat(
        clumps_img,
        variables,
        sk_classifier,
        cls_info_dict,
        out_col_int="OutClass",
        out_col_str="OutClassName",
        roi_col=None,
        roi_val=1,
        class_colours=True,
    )

    read_out_cls = False
    try:
        cls_col_vals = rsgislib.rastergis.get_column_data(clumps_img, "OutClass")
        read_out_cls = True
    except:
        read_out_cls = False

    assert read_out_cls



#!/usr/bin/env python
"""
These utilities allow for easier use of the Look Up Table (LUT) functionality within
RSGISLib.
"""
############################################################################
#  imagelut.py
#
#  Copyright 2019 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide a set of utilities for working with image LUTs in RSGISLib
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 23/10/2019
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
############################################################################

import os
from typing import List

from osgeo import gdal
from osgeo import ogr

gdal.UseExceptions()

import rsgislib
import rsgislib.vectorutils


def create_img_extent_lut(
    input_imgs: List[str],
    vec_file: str,
    vec_lyr: str,
    out_format: str,
    ignore_none_imgs: bool = False,
    out_proj_wgs84: bool = False,
    overwrite_lut_file: bool = False,
):
    """
    Create a vector layer look up table (LUT) for a directory of images.

    :param input_imgs: list of input images for the LUT. All input images should be
                       the same projection/coordinate system.
    :param vec_file: output vector file/path
    :param vec_lyr: output vector layer
    :param out_format: the output vector layer type (e.g., GPKG).
    :param ignore_none_imgs: if a NULL epsg is returned from an image then ignore
                             and don't include in LUT else throw exception.
    :param out_proj_wgs84: if True then the image bounding boxes will be
                           re-projected to EPSG:4326.
    :param overwrite_lut_file: if True then output file will be overwritten. If
                               false then not, e.g., can add extra layer to GPKG

    .. code:: python

        import glob
        import rsgislib.imageutils.imagelut
        imgList = glob.glob('/Users/pete/Temp/GabonLandsat/Hansen*.kea')
        rsgislib.imageutils.imagelut.create_img_extent_lut(imgList,
                                                          'ImgExtents.gpkg',
                                                          'HansenImgExtents', 'GPKG')

    """
    import tqdm
    import rsgislib.imageutils
    import rsgislib.vectorutils.createvectors

    bboxs = []
    atts = dict()
    atts["filename"] = []
    atts["path"] = []

    attTypes = dict()
    attTypes["types"] = [ogr.OFTString, ogr.OFTString]
    attTypes["names"] = ["filename", "path"]

    epsgCode = 0

    first = True
    baseImg = ""
    for img in tqdm.tqdm(input_imgs):
        epsgCodeTmp = rsgislib.imageutils.get_epsg_proj_from_img(img)
        epsg_found = True
        if epsgCodeTmp is None:
            epsg_found = False
            if not ignore_none_imgs:
                raise rsgislib.RSGISPyException(
                    "The EPSG code is None: '{}'".format(img)
                )
        if epsg_found:
            if out_proj_wgs84:
                epsgCode = 4326
            else:
                epsgCodeTmp = int(epsgCodeTmp)
                if first:
                    epsgCode = epsgCodeTmp
                    baseImg = img
                    first = False
                else:
                    if epsgCodeTmp != epsgCode:
                        raise rsgislib.RSGISPyException(
                            "The EPSG codes ({0} & {1}) do not match. "
                            "(Base: '{2}', Img: '{3}')".format(
                                epsgCode, epsgCodeTmp, baseImg, img
                            )
                        )

            if out_proj_wgs84:
                img_bbox = rsgislib.imageutils.get_img_bbox_in_proj(img, 4326)
            else:
                img_bbox = rsgislib.imageutils.get_img_bbox(img)

            bboxs.append(img_bbox)
            baseName = os.path.basename(img)
            filePath = os.path.dirname(img)
            atts["filename"].append(baseName)
            atts["path"].append(filePath)
    # Create vector layer
    rsgislib.vectorutils.createvectors.create_poly_vec_bboxs(
        vec_file,
        vec_lyr,
        out_format,
        epsgCode,
        bboxs,
        atts,
        attTypes,
        overwrite=overwrite_lut_file,
    )


def query_img_lut(scn_bbox: List[float], lut_db_file: str, lyr_name: str) -> List[str]:
    """
    A function for querying the LUT DB spatially filtering using a BBOX

    :param scn_bbox: A bbox (MinX, MaxX, MinY, MaxY) in the same projection as
                     the LUT for the area of interest.
    :param lut_db_file: The file path to the LUT database file (e.g., lut.gpkg).
    :param lyr_name: The layer name within the database file.
    :return: a list of files from the LUT

    """
    file_list_lut = rsgislib.vectorutils.get_att_lst_select_bbox_feats(
        lut_db_file, lyr_name, ["path", "filename"], scn_bbox
    )
    imgs = []
    for item in file_list_lut:
        imgs.append(os.path.join(item["path"], item["filename"]))
    return imgs


def get_all_lut_imgs(lut_db_file: str, lyr_name: str) -> List[str]:
    """
    Get a list of all the images within the LUT.

    :param lut_db_file: The file path to the LUT database file (e.g., lut.gpkg).
    :param lyr_name: The layer name within the database file.
    :return: a list of files from the LUT

    """
    import rsgislib.vectorattrs

    file_list_lut = rsgislib.vectorattrs.read_vec_columns(
        lut_db_file, lyr_name, ["path", "filename"]
    )
    imgs = []
    for item in file_list_lut:
        imgs.append(os.path.join(item["path"], item["filename"]))
    return imgs


def get_raster_lyr(
    scn_bbox: List[float], lut_db_file: str, lyr_name: str, tmp_dir: str
) -> str:
    """
    This function provides a single raster layer using the LUT file provided. If
    a single image file intersecting with the BBOX is not within the LUT then a
    VRT generated from a number of images will be returned. The VRT will be within
    the tmp_path directory which will need to be cleaned up afterwards. If no image(s)
    is available then None is returned.

    :param scn_bbox: A bbox (MinX, MaxX, MinY, MaxY) in the same projection as the
                     LUT for the area of interest.
    :param lut_db_file: The file path to the LUT datbase file (e.g., lut.gpkg).
    :param lyr_name: The layer name within the database file.
    :param tmp_dir: A directory where temporary files can be saved to.
    :return: a string with the file path to the raster layer. If no image intersects
             then None is returned.

    """
    import rsgislib.tools.utils

    file_list_lut = rsgislib.vectorutils.get_att_lst_select_bbox_feats(
        lut_db_file, lyr_name, ["path", "filename"], scn_bbox
    )

    img_base = "imglyr_{}".format(rsgislib.tools.utils.uid_generator())
    # if number of scenes available is > 0 then create VRT
    if len(file_list_lut) > 1:
        imgs = []
        for item in file_list_lut:
            imgs.append(os.path.join(item["path"], item["filename"]))
        out_img_file = os.path.join(tmp_dir, "{}_tmp.vrt".format(img_base))
        gdal.BuildVRT(out_img_file, imgs)
    elif len(file_list_lut) == 1:
        out_img_file = os.path.join(
            file_list_lut[0]["path"], file_list_lut[0]["filename"]
        )
    else:
        out_img_file = None
    return out_img_file


def query_file_lut(
    lut_db_file: str,
    lyr_name: str,
    roi_file: str,
    roi_lyr: str,
    out_dest: str,
    targz_out: bool,
    cp_cmds: bool,
) -> List[str]:
    """
    A function which allows the file LUT to be queried (intersection) and commands
    generated for completing operations. Must select (pass true) for either targz_out
    or cp_cmds not both. If both are False then the list of intersecting files will
    be returned.

    :param lut_db_file: OGR vector file with the LUT.
    :param lyr_name: name of the layer within the LUT file.
    :param roi_file: region of interest OGR vector file.
    :param roi_lyr: layer name within the ROI file.
    :param out_dest: the destination for outputs from command (e.g., where are the
                     files to be copied to or output file name for tar.gz file.
    :param targz_out: boolean which specifies that the command for generating
                      a tar.gz file should be generated.
    :param cp_cmds: boolean which specifies that the command for copying the
                    LUT files to a out_dest should be generated.

    :return: returns a list of commands to be executed.

    """
    if lyr_name is None:
        lyr_name = os.path.splitext(os.path.basename(lut_db_file))[0]

    if roi_lyr is None:
        roi_lyr = os.path.splitext(os.path.basename(roi_file))[0]

    roi_mem_ds, roi_mem_lyr = rsgislib.vectorutils.read_vec_lyr_to_mem(
        roi_file, roi_lyr
    )

    roi_bbox = roi_mem_lyr.GetExtent(True)

    lut_mem_ds, lut_mem_lyr = rsgislib.vectorutils.get_mem_vec_lyr_subset(
        lut_db_file, lyr_name, roi_bbox
    )

    file_list_dict = rsgislib.vectorutils.get_att_lst_select_feats_lyr_objs(
        lut_mem_lyr, ["path", "filename"], roi_mem_lyr
    )

    out_cmds = []
    if targz_out:
        cmd = "tar -czf {}".format(out_dest)
        for file_item in file_list_dict:
            file_path = os.path.join(file_item["path"], file_item["filename"])
            cmd = "{} {}".format(cmd, file_path)
        out_cmds.append(cmd)
    elif cp_cmds:
        for file_item in file_list_dict:
            file_path = os.path.join(file_item["path"], file_item["filename"])
            out_cmds.append("cp {0} {1}".format(file_path, out_dest))
    else:
        for file_item in file_list_dict:
            file_path = os.path.join(file_item["path"], file_item["filename"])
            out_cmds.append(file_path)

    return out_cmds



import os
import pytest

SKLEARN_NOT_AVAIL = False
try:
    import sklearn
except ImportError:
    SKLEARN_NOT_AVAIL = True

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")


@pytest.mark.skipif(SKLEARN_NOT_AVAIL, reason="scikit-learn dependency not available")
def test_img_pixel_sample_cluster(tmp_path):
    import rsgislib.classification.clustersklearn

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_img = os.path.join(tmp_path, "sen2_20210527_aber_clustered.kea")
    rsgislib.classification.clustersklearn.img_pixel_sample_cluster(
        input_img, output_img, gdalformat="KEA"
    )
    assert os.path.exists(output_img)


@pytest.mark.skipif(SKLEARN_NOT_AVAIL, reason="scikit-learn dependency not available")
def test_img_pixel_tiled_cluster(tmp_path):
    import rsgislib.classification.clustersklearn

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_img = os.path.join(tmp_path, "sen2_20210527_aber_clustered.kea")
    rsgislib.classification.clustersklearn.img_pixel_tiled_cluster(
        input_img, output_img, gdalformat="KEA"
    )
    assert os.path.exists(output_img)


@pytest.mark.skipif(SKLEARN_NOT_AVAIL, reason="scikit-learn dependency not available")
def test_img_pixel_cluster(tmp_path):
    import rsgislib.classification.clustersklearn

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_img = os.path.join(tmp_path, "sen2_20210527_aber_clustered.kea")
    rsgislib.classification.clustersklearn.img_pixel_cluster(
        input_img, output_img, gdalformat="KEA"
    )
    assert os.path.exists(output_img)



#!/usr/bin/env python

from sklearn.ensemble import ExtraTreesRegressor
from sklearn.model_selection import GridSearchCV

import rsgislib


def get_ann_obj_params(n_predictors):
    """
    Get a Artificial neural network object and parameters.

    :return: set [ANN Object, ANN Parameters Dict, Boolean as to whether data needs scaling]

    """
    import numpy
    from sklearn.neural_network import MLPRegressor

    ann_grid = {
        "hidden_layer_sizes": [(10,), (25,), (50,), (100,)],
        "learning_rate": ["constant", "adaptive"],
        "learning_rate_init": numpy.arange(0.01, 0.1, 0.005).tolist(),
        "alpha": numpy.arange(0.01, 0.3, 0.002).tolist(),
        "max_iter": numpy.arange(300, 1000, 100, dtype="uint16").tolist(),
    }
    ann_obj = MLPRegressor(solver="lbfgs", activation="relu")
    return ann_obj, ann_grid, True


def get_en_obj_params(n_predictors):
    """
    Get a ElasticNet object and parameters.

    :return: set [EN Object, EN Parameters Dict, Boolean as to whether data needs scaling]

    """
    import numpy
    from sklearn.linear_model import ElasticNet

    en_grid = {
        "alpha": numpy.arange(1.0, 5.0, 0.1).tolist(),
        "l1_ratio": numpy.arange(0.01, 1.0, 0.01).tolist(),
    }
    en_obj = ElasticNet(max_iter=10000, tol=0.01)
    return en_obj, en_grid, False


def get_knn_obj_params(n_predictors):
    """
    Get a KNeighborsRegressor object and parameters.

    :return: set [KNN Object, KNN Parameters Dict, Boolean as to whether data needs scaling]

    """
    import numpy
    from sklearn.neighbors import KNeighborsRegressor

    knn_grid = {
        "n_neighbors": numpy.arange(1, 9, 1, dtype="uint8").tolist(),
        "weights": ["uniform", "distance"],
    }
    knn_obj = KNeighborsRegressor(algorithm="brute")
    return knn_obj, knn_grid, True


def get_kr_obj_params(n_predictors):
    """
    Get a KernelRidge object and parameters.

    :return: set [KR Object, KR Parameters Dict, Boolean as to whether data needs scaling]

    """
    import numpy
    from sklearn.kernel_ridge import KernelRidge

    kr_grid = {
        "alpha": numpy.arange(0.05, 2.0, 0.05).tolist(),
        "gamma": numpy.arange(0.1, 2.5, 0.1).tolist(),
        "kernel": ["rbf", "linear", "sigmoid"],
    }
    kr_obj = KernelRidge()
    return kr_obj, kr_grid, False


def get_et_obj_params(n_predictors):
    """
    Get a ExtraTreesRegressor object and parameters.

    :return: set [ET Object, ET Parameters Dict, Boolean as to whether data needs scaling]

    """
    import numpy
    from sklearn.ensemble import ExtraTreesRegressor

    et_grid = {
        "n_estimators": numpy.arange(100, 600, 100, dtype="uint16").tolist(),
        "max_features": numpy.arange(1, n_predictors, 1, dtype="uint8").tolist(),
        "max_depth": numpy.arange(1, n_predictors, 1, dtype="uint8").tolist(),
        "min_samples_split": numpy.arange(2, 8, 1, dtype="uint8").tolist(),
        "min_samples_leaf": numpy.arange(1, 6, 1, dtype="uint8").tolist(),
    }
    et_obj = ExtraTreesRegressor()
    return et_obj, et_grid, False


def get_pls_obj_params(n_predictors):
    """
    Get a PLSRegression object and parameters.

    :return: set [PLS Object, PLS Parameters Dict, Boolean as to whether data needs scaling]

    """
    import numpy
    from sklearn.cross_decomposition import PLSRegression

    pls_grid = {
        "n_components": numpy.arange(1, n_predictors + 1, 1, dtype="uint8").tolist()
    }
    pls_obj = PLSRegression(scale=False, max_iter=1000, tol=0.001, copy=True)
    return pls_obj, pls_grid, False


def create_search_obj(regrs_obj, regrs_params, n_runs=250, n_cv=5, n_cores=1):
    """
    A function which creates a scikit-learn search object (i.e., GridSearchCV or
    RandomizedSearchCV) to be used within the perform_search_param_opt function
    to identify the optimal algorithms for the algorithm.

    Default is to use a Grid Search which tries all combinations but if that
    is too many runs then a random search is used.

    :param regrs_obj: The scikit-learn object (e.g., ExtraTreesRegressor)
    :param regrs_params: a dict of the parameters to search. i.e., provided by
                         get_XX_obj_params functions (e.g., get_et_obj_params)
    :param n_runs: The maximum number of runs to perform. If the required number of
                   runs is > n_runs then RandomizedSearchCV is used.
    :param n_cv: the number of cross-validations to use for the analysis
    :param n_cores: the number of cores to use.
    :return: the instance of a scikit-learn BaseSearchCV (i.e., either GridSearchCV
             or RandomizedSearchCV)

    """
    from sklearn.model_selection import ParameterGrid
    from sklearn.model_selection import RandomizedSearchCV
    from sklearn.model_selection import GridSearchCV

    if len(ParameterGrid(regrs_params)) > n_runs:
        skl_srch_obj = GridSearchCV(
            regrs_obj,
            regrs_params,
            scoring="neg_mean_squared_error",
            cv=n_cv,
            n_jobs=n_cores,
            verbose=1,
        )
    else:
        skl_srch_obj = RandomizedSearchCV(
            regrs_obj,
            regrs_params,
            scoring="neg_mean_squared_error",
            cv=n_cv,
            n_iter=n_runs,
            n_jobs=n_cores,
            verbose=1,
        )

    return skl_srch_obj


def perform_search_param_opt(
    opt_params_file, x, y, skl_srch_obj=None, data_scaler=None
):
    """
    A function which performs a parameter optimisation using an instance of a
    scikit-learn BaseSearchCV (i.e., either GridSearchCV or RandomizedSearchCV).

    :param opt_params_file: output JSON file where there optimal parameters
                            will be written.
    :param x: The independent variables used to estimate y.
    :param y: The dependent variable(s) to for the which the regression will be
              carried out.
    :param skl_srch_obj: a scikit-learn BaseSearchCV (i.e., either GridSearchCV or
                         RandomizedSearchCV) instance.
    :param data_scaler: optional data scaler from scikit learn
    :return: the optimal scikit learn regression object.

    """
    import json
    import numpy

    if data_scaler is not None:
        x_train = data_scaler.transform(x)
    else:
        x_train = numpy.copy(x)

    skl_srch_obj.fit(x_train, y)

    best_params = skl_srch_obj.best_params_

    with open(opt_params_file, "w") as fp:
        json.dump(
            best_params,
            fp,
            sort_keys=True,
            indent=4,
            separators=(",", ": "),
            ensure_ascii=False,
        )

    return skl_srch_obj.best_estimator_


def perform_kfold_fit(
    skl_regrs_obj, x, y, n_splits=5, repeats=1, shuffle=False, data_scaler=None
):
    """
    A function which performs a k-fold fitting of a regression model to a dataset
    to estimate the output model variance in the relevant summary metrics.

    :param skl_regrs_obj: A scikit learn regression object.
    :param x: The independent variables used to estimate y.
    :param y: The dependent variable(s) to for the which the regression will be
              carried out.
    :param n_splits: number of splits to perform
    :param repeats: the number of times to repeat each split.
    :param shuffle: if not using repeats and shuffle=True then the data will be
                    shuffled before splitting.
    :param data_scaler: optional data scaler from scikit learn
    :return: acc_metrics, residuals

    """
    from sklearn.model_selection import KFold, RepeatedKFold
    import sklearn.metrics
    import numpy
    import rsgislib.tools.stats
    import tqdm

    if repeats > 1:
        kf = RepeatedKFold(n_splits=n_splits, n_repeats=repeats)
    else:
        kf = KFold(n_splits=n_splits, shuffle=shuffle)

    y_true = list()
    y_pred = list()

    acc_metrics = []
    n_dep_var = y.shape[1]
    for i in range(n_dep_var):
        acc_metrics_dict = dict()
        acc_metrics_dict["r2"] = list()
        acc_metrics_dict["explained_variance_score"] = list()
        acc_metrics_dict["median_absolute_error"] = list()
        acc_metrics_dict["mean_absolute_error"] = list()
        acc_metrics_dict["mean_squared_error"] = list()
        acc_metrics_dict["root_mean_squared_error"] = list()
        acc_metrics_dict["norm_root_mean_squared_error"] = list()
        acc_metrics_dict["bias"] = list()
        acc_metrics_dict["norm_bias"] = list()
        acc_metrics_dict["bias_squared"] = list()
        acc_metrics_dict["variance"] = list()
        acc_metrics_dict["noise"] = list()
        acc_metrics.append(acc_metrics_dict)

    for train_idx, test_idx in tqdm.tqdm(kf.split(x)):
        x_train, x_test = x[train_idx], x[test_idx]
        y_train, y_test = y[train_idx], y[test_idx]
        y_true.append(y_test)

        if data_scaler is not None:
            x_train = data_scaler.transform(x_train)
            x_test = data_scaler.transform(x_test)

        skl_regrs_obj.fit(x_train, y_train)

        y_hat = skl_regrs_obj.predict(x_test)

        if n_dep_var == 1:
            y_hat = numpy.expand_dims(y_hat, axis=1)

        for i in range(n_dep_var):
            acc_metrics[i]["r2"].append(
                sklearn.metrics.r2_score(y_test[..., i], y_hat[..., i])
            )
            acc_metrics[i]["explained_variance_score"].append(
                sklearn.metrics.explained_variance_score(y_test[..., i], y_hat[..., i])
            )
            acc_metrics[i]["median_absolute_error"].append(
                sklearn.metrics.median_absolute_error(y_test[..., i], y_hat[..., i])
            )
            acc_metrics[i]["mean_absolute_error"].append(
                sklearn.metrics.mean_absolute_error(y_test[..., i], y_hat[..., i])
            )
            acc_metrics[i]["mean_squared_error"].append(
                sklearn.metrics.mean_squared_error(y_test[..., i], y_hat[..., i])
            )
            rmse = numpy.sqrt(
                sklearn.metrics.mean_squared_error(y_test[..., i], y_hat[..., i])
            )
            acc_metrics[i]["root_mean_squared_error"].append(rmse)
            acc_metrics[i]["norm_root_mean_squared_error"].append(
                100 * (rmse / numpy.mean(y_test[..., i]))
            )
            bias, norm_bias = rsgislib.tools.stats.bias_score(
                y_test[..., i], y_hat[..., i]
            )
            acc_metrics[i]["bias"].append(bias)
            acc_metrics[i]["norm_bias"].append(norm_bias)
            (
                mse,
                bias_squared,
                variance,
                noise,
            ) = rsgislib.tools.stats.decompose_bias_variance(
                y_test[..., i], y_hat[..., i]
            )
            acc_metrics[i]["bias_squared"].append(bias_squared)
            acc_metrics[i]["variance"].append(variance)
            acc_metrics[i]["noise"].append(noise)

        y_pred.append(y_hat)

    # Convert the results into np.ndarrays, shape=(n_responses, n_samples)
    y_true = numpy.concatenate(y_true).T
    y_pred = numpy.concatenate(y_pred).T

    # Save rediduals within a list of dicts
    residuals = []
    for i in range(n_dep_var):
        c_residuals = dict()
        c_residuals["y_true"] = y_true[i]
        c_residuals["y_pred"] = y_pred[i]
        residuals.append(c_residuals)

    return acc_metrics, residuals


def apply_regress_sklearn_mdl(
    regrs_mdl,
    n_out_vars,
    predictor_img,
    predictor_img_bands,
    vld_msk_img,
    vld_msk_val,
    out_img,
    gdalformat="KEA",
    out_band_names=None,
    calc_stats=True,
    out_no_date_val=0.0,
):
    """

    :param regrs_mdl: the scikit-learn model
    :param n_out_vars: the number of output variables (i.e., image bands)
    :param predictor_img: the input image file providing the independent predictor
                          variables used as the inputs to the model.
    :param predictor_img_bands: list of the image bands used. Ensure this is in the
                                same order as the variables used to train the model.
    :param vld_msk_img: An input image file defining where the model should be applied
    :param vld_msk_val: the pixel value within the vld_msk_img specifying the pixels
                        to which the model should be applied to.
    :param out_img: the output image file path.
    :param gdalformat: output image file format (Default: KEA)
    :param out_band_names: Optional list of band names for the output image. If None
                           (Default) then not band names will be defined.
    :param calc_stats: boolean specifying whether image statistics and pyramids
                       should be build (default: True)
    :param out_no_date_val: Output no data value to be used within the image file.

    """

    from rios import applier
    from rios import cuiprogress
    import numpy
    import rsgislib.imageutils

    try:
        import tqdm

        progress_bar = rsgislib.TQDMProgressBar()
    except:
        progress_bar = cuiprogress.GDALProgressBar()

    # Convert from GDAL band index (start 1) to Numpy Array index (start 0)
    predictor_img_bands_arr = numpy.array(predictor_img_bands)
    predictor_img_bands_arr = predictor_img_bands_arr - 1
    if numpy.min(predictor_img_bands_arr) < 0:
        raise rsgislib.RSGISPyException(
            "Image band numbering starts at 1; i.e., GDAL convension."
        )

    infiles = applier.FilenameAssociations()
    infiles.vld_msk_img = vld_msk_img
    infiles.predictor_img = predictor_img

    outfiles = applier.FilenameAssociations()
    outfiles.out_img = out_img

    otherargs = applier.OtherInputs()
    otherargs.regrs_mdl = regrs_mdl
    otherargs.vld_msk_val = vld_msk_val
    otherargs.predictor_img_bands = predictor_img_bands_arr
    otherargs.n_out_vars = n_out_vars
    otherargs.out_no_date_val = out_no_date_val

    aControls = applier.ApplierControls()
    aControls.progress = progress_bar
    aControls.drivername = gdalformat
    aControls.omitPyramids = True
    aControls.calcStats = False

    def _applySKRegressionModel(info, inputs, outputs, otherargs):
        """
        Internal function to apply scikit-learn regression model.
        """
        if numpy.any(inputs.vld_msk_img == otherargs.vld_msk_val):
            # Flatten the input image.
            pred_flat_data = numpy.moveaxis(inputs.predictor_img, 0, 2).reshape(
                -1, inputs.predictor_img.shape[0]
            )
            # Subset to the relevant metrics.
            pred_flat_data = pred_flat_data.take(otherargs.predictor_img_bands, axis=1)

            ID = numpy.arange(pred_flat_data.shape[0])
            n_feats = ID.shape[0]

            # Mask to the valid data regions
            ID = ID[inputs.vld_msk_img.flatten() == otherargs.vld_msk_val]
            pred_flat_data = pred_flat_data[
                inputs.vld_msk_img.flatten() == otherargs.vld_msk_val
            ]

            # Run the model
            mdl_est_vals = otherargs.regrs_mdl.predict(pred_flat_data)
            if otherargs.n_out_vars == 1:
                mdl_est_vals = numpy.expand_dims(mdl_est_vals, axis=1)

            # Create the output data array
            out_vals = numpy.zeros([n_feats, otherargs.n_out_vars])
            if otherargs.out_no_date_val != 0.0:
                out_vals[...] = otherargs.out_no_date_val

            # Copy the model outputs to the full array
            out_vals[ID] = mdl_est_vals

            # Reshape the output array.
            out_img_vals = out_vals.reshape(
                inputs.vld_msk_img.shape[1],
                inputs.vld_msk_img.shape[2],
                otherargs.n_out_vars,
            )
            out_img_vals = numpy.moveaxis(out_img_vals, 2, 0)
            outputs.out_img = out_img_vals
        else:
            out_img_vals = numpy.zeros(
                [
                    otherargs.n_out_vars,
                    inputs.vld_msk_img.shape[1],
                    inputs.vld_msk_img.shape[2],
                ],
                dtype=numpy.float32,
            )
            if otherargs.out_no_date_val != 0.0:
                out_img_vals[...] = otherargs.out_no_date_val
            outputs.out_img = out_img_vals

    print("Applying the Regression model")
    applier.apply(
        _applySKRegressionModel, infiles, outfiles, otherargs, controls=aControls
    )
    print("Completed")

    if out_band_names is not None:
        if n_out_vars == len(out_band_names):
            rsgislib.imageutils.set_band_names(out_img, out_band_names)
        else:
            print(
                "The number of output variables and the number of bands names provided do not match so ignoring."
            )

    if calc_stats:
        rsgislib.imageutils.pop_img_stats(
            out_img, use_no_data=True, no_data_val=out_no_date_val, calc_pyramids=True
        )



#!/usr/bin/env python

#############################################
# rsgistranslate2tif.py
#
#  Copyright 2020 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
# Purpose:  Using gdal_translate converts a single input image to GEOTIFF
#           tiling, compressing and copying any image overviews.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 29/05/2020
# Version: 1.0
#
#############################################

import argparse
import subprocess

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", type=str, required=True, help="The input image file")
    parser.add_argument("-o", "--output", type=str, required=True, help="The output image (GTIFF) file.")
    args = parser.parse_args()

    cmd = "gdal_translate -of GTIFF -co TILED=YES -co COMPRESS=LZW -co " \
          "BIGTIFF=IF_SAFER -co COPY_SRC_OVERVIEWS=YES  {} {}".format(args.input, args.output)
    print("Running: '{}'".format(cmd))
    try:
        subprocess.check_call(cmd, shell=True)
    except OSError as e:
        raise Exception('Could not execute command: {}'.format(cmd))




#!/usr/bin/env python
"""
The tools and utilities for specific sensors.
"""


def read_landsat_mtl_to_dict(in_mtl_file):
    """
    A function which can read a Landsat MTL header file and return a
    structured dict with the information from the MTL file.

    :param in_mtl_file: the file path to MTL header file (*MTL.txt)
    :return: dict structure.

    .. code:: python

        import pprint
        import rsgislib.tools.sensors

        ls_header = rsgislib.tools.sensors.read_landsat_mtl_to_dict("landsat_MTL.txt")
        pprint.pprint(ls_header)

    """
    h_file = open(in_mtl_file, "r")
    header_params = dict()
    for line in h_file:
        line = line.strip()
        if line:
            line_vals = line.split("=")
            if len(line_vals) == 2:
                if line_vals[0].strip() == "GROUP":
                    c_dict = dict()
                    header_params[line_vals[1].strip()] = c_dict
                elif line_vals[0].strip() != "END_GROUP":
                    c_dict[line_vals[0].strip()] = line_vals[1].strip().replace('"', "")
    h_file.close()
    return header_params
